/**
 * 
 */
package compactTriangulations;

import arraybasedDS.NavigationInterface;
import arraybasedDS.WingededgeInterface;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique)
 *
 * Abstract class defining the methods provided by Compact Data Structures for triangle meshes
 */
public abstract class CompactTriangleMesh implements NavigationInterface, WingededgeInterface {

	public int v0, v1, v2; // (indices of) vertices on the root face (used to handle special cases)
	public int e10, e20, e21; // (indices of) edges on the root face (used to handle special cases)
	protected int nVertices;
	public float[] coordF; // geometric coordinates (for testing performances without java interfaces)
	
	/** 
	 * Return the 'LeftFront' edge, neighbor of 'e'
	 * @param e an edge
	 * @return the 'LeftFront' edge incident to 'e'
	 **/
	public abstract int getLeftFront(int e);
	
	/** 
	 * Return the 'RightFront' edge, neighbor of 'e'
	 * @param e an edge
	 * @return the 'RightFront' edge incident to 'e'
	 **/
	public abstract int getRightFront(int e);

	/** 
	 * Return the 'LeftBack' edge, neighbor of 'e'
	 * @param e an edge
	 * @return the 'LeftBack' edge incident to 'e'
	 **/
	public abstract int getLeftBack(int e);

	/** 
	 * Return the 'RightBack' edge, neighbor of 'e'
	 * @param e an edge
	 * @return the 'RightBack' edge incident to 'e'
	 **/
	public abstract int getRightBack(int e);

	/** 
	 * Return the vertex source of 'e'
	 * @param e  an edge
	 **/
	public abstract int getSource(int e);
	
	/** 
	 * Return the vertex target of 'e'
	 * @param e  an edge
	 **/
	public abstract int getTarget(int e);
	
	/** 
	 * Return an edge incident to vertex v
	 * @param v  the index of the vertex
	 **/
	public abstract int getEdge(int v);	
	/** 
	 * Check whether the edge 'e' is outgoing from vertex 'v'
	 * 
	 * @param e  an edge
	 * @param v  the index of a vertex
	 * @return return true whether 'e' is outgoing from v
	 **/
	public abstract boolean isOutgoing(int e, int v);
	
	/**
	 * Return c+1 (mod 3)
	 */
	public int nextColor(int color) {
		if(color==0) return 1;
		if(color==1) return 2;
		return 0;
	}
	
	/**
	 * Return c-1 (mod 3)
	 */
	public int prevColor(int color) {
		if(color==2) return 1;
		if(color==1) return 0;
		return 2;
	}
	
	/** 
	 * Return the geometric coordinates of vertex v
	 * 
	 * @param v  the index of the vertex
	 **/
	public float[] getPoint(int v) {
		int index=3*v;
		return new float[] {this.coordF[index], this.coordF[index+1], this.coordF[index+2]};
	}

	/** 
	 * Set the geometric coordinates of vertex v
	 * 
	 * @param v  the index of the vertex
	 **/
	public void setPoint(int v, float x, float y, float z) {
		int index=3*v;
		this.coordF[index]=x;
		this.coordF[index+1]=y;
		this.coordF[index+2]=z;
	}

	/** Return the x-coordinate of vertex v **/
	public float getX(int v) { return this.coordF[v*3]; }

	/** Return the y-coordinate of vertex v **/
	public float getY(int v) { return this.coordF[v*3+1]; }

	/** Return the z-coordinate of vertex v **/
	public float getZ(int v) { return this.coordF[v*3+2]; }

	public String vectorToString(float[] u) {
		  return "("+u[0]+","+u[1]+","+u[2]+")";
	}

}
