package compactTriangulations;

import arrays.*;
import arraybasedDS.*;

/**
 * @author Luca Castelli Aleardi
 * 
 * Array based implementation of a compact data structure for triangle meshes,
 * requiring at most 4n references.
 * Use of 4 service bits
 */
public class CompactTriangleMesh_4n extends CompactTriangleMesh {
	protected TableWithFourServiceBits table;
	
	private final int nReferences=4; // number of references per vertex
	private int nVertices;
	private int nEdges;
	private int tableSize; // total number of entries in the table
	
	public CompactTriangleMesh_4n(int n) {
		this.nVertices=n;
		this.nEdges=3*n;
		this.tableSize=4*n;
		this.table=new TableWithFourServiceBits(tableSize);
		
		//this.coordF=new float[nVertices*3]; // geometric coordinates
		this.coordF=null; // the coordinates will be initialized during the construction/decompression phase
	}
	
	//--- methods for initializing integer references -------
	
	/**
	 * Store the index of left neighbor of edge e
	 */
	public void setLeft(int e, int eLeft) {
		int v=e/3;
		int eColor=e%3;
		if(eColor==0) this.table.set(4*v, eLeft);
		else if(eColor==1) this.table.set(4*v+1, eLeft);
		else this.table.set(4*v+3, eLeft);
	}
	
	/**
	 * Store the index of right neighbor of edge e
	 */
	public void setRight(int e, int eRight) {
		int v=e/3;
		int eColor=e%3;
		if(eColor==0) this.table.set(4*v, eRight);
		else if(eColor==1) this.table.set(4*v+2, eRight);
		else this.table.set(4*v+3, eRight);
	}
	
	//--- methods for initializing service bits --------
	// every reference stores 4 service bits, describing orientations of neighboring edges
	
	/**
	 * Store the first service bit of an edge: it corresponds to ...
	 * The bit is set to 1.
	 */
	public void setLeftBit0(int e) { 
		int v=e/3;
		int eColor=e%3;
		if(eColor==0) this.table.setBit0(4*v);
		else if(eColor==1) this.table.setBit0(4*v+1);
		else this.table.setBit0(4*v+3);
	}
	
	/**
	 * Store the second service bit of an edge: it corresponds to ...
	 * The bit is set to 1.
	 */
	public void setLeftBit1(int e) { 
		int v=e/3;
		int eColor=e%3;
		if(eColor==0) this.table.setBit1(4*v);
		else if(eColor==1) this.table.setBit1(4*v+1);
		else this.table.setBit1(4*v+3);
	}
	
	/**
	 * Store the third service bit of an edge: it corresponds to ...
	 * The bit is set to 1.
	 */
	public void setRightBit0(int e) {
		int v=e/3;
		int eColor=e%3;
		if(eColor==0) this.table.setBit2(4*v);
		else if(eColor==1) this.table.setBit2(4*v+1);
		else this.table.setBit2(4*v+3);
	}

	/**
	 * Store the forth service bit of an edge: it corresponds to ...
	 * The bit is set to 1.
	 */
	public void setRightBit1(int e) {
		int v=e/3;
		int eColor=e%3;
		if(eColor==0) this.table.setBit3(4*v);
		else if(eColor==1) this.table.setBit3(4*v+1);
		else this.table.setBit3(4*v+3);
	}

	//--- methods for accessing information and navigating between mesh elements ---
	
	/**
	 * Return the number of vertices
	 */
	public int sizeOfVertices() { return this.nVertices; }

	/**
	 * Return the number of edges
	 */
	public int sizeOfEdges() { return this.nEdges; }
	
	/**
	 * Perform getLeftFront operator: return the index of edge vw, neighbor of edge e=uv
	 */
	public int getLeftFront(int e) {
		if(e==e10) return e20;
		if(e==e21) return e10;
		
		int u=e/3;
		int eColor=e%3;
		
		if(eColor==0)  {// red edges
			int bitLeft=this.table.getBit1(4*u); // orientation of edge vw
			if(bitLeft==1) // edge vw is oriented toward v
				return e-3; // because of ordering on red edges (e.g. BFS or DFUDS ordering)
			else
				return this.table.elementAt(4*u+2); // since vw=getRightFront(uw)
		}

		if(eColor==1) // blue edges
			return this.table.elementAt(4*u+1);
		
		else {// black edges
			int bitLeft=this.table.getBit0(4*u+3); // orientation of edge uw
			int bitRight=this.table.getBit2(4*u+3); // orientation of edge uz
			if(bitRight==0) // edge (u,z) is outgoing from u
				return this.table.elementAt(4*u+3);
			if(bitLeft==0) // edge (u,w) is outgoing from u
				return this.getRightFront(e-2); // e-2 is of color 0, incident to v
			else { // edge (u,w) is oriented toward u
				int eRightBack=this.getRightBack(e); // edge uz is red and oriented toward u
				int z=eRightBack/3;
				int eLeftBack=this.table.elementAt(4*z); // edge uz stores explicitly its neighbor
				return eLeftBack+1;
			}
		}

	}

	/**
	 * Perform getRightFront operator: return the index of edge vz, neighbor of edge e=uv
	 */
	public int getRightFront(int e) {
		//System.out.print("RightFront "+e);
		if(e==e20) return e10; // edge e belongs to the root (exterior) face
		
		int u=e/3;
		int eColor=e%3;
			
		if(eColor==0) { // red edges
			int r=4*u;
			int bitRight=this.table.getBit3(r); // orientation of edge vz
			if(bitRight==1) // edge vz is oriented toward v
				return e+3; // because of ordering on red edges (e.g. BFS or DFUDS ordering)

			int rightReference=this.table.elementAt(r);
			int rightColor=rightReference%3;
			if(rightColor!=1) // easy case: the reference is stored
				return rightReference;
			else return this.getLeftFront(rightReference);
		}
		else if(eColor==1) { // blue edges (easy case)
			return this.table.elementAt(4*u+2);
		}
		else { // black edges
			int bit=this.table.getBit2(4*u+3); // orientation of edge uz
			if(bit==1) // edge (u,z) is oriented toward u, of color 0
				return this.table.elementAt(4*u+3);
			else
				return this.getLeftFront(e-1); // edge uz must be of color 1
		}
		
	}

	// LCA: seems OK
	public int getLeftBack(int e) {
		throw new Error("a' completer");
		/*
		//System.out.print("leftBack "+e);
		if(e==e10)
			return e21;
		if(e==e21)
			return e20;
		int edgeColor=e%3;
		// the leftBack edge is oriented outgoing from v
		if(this.table.getFirstServiceBit(2*e)==0) {
			if(edgeColor==2) return e-2;
			return e+1;
		}
		// the leftBack edge is oriented toward v
		else {
			int eLeftFront=this.table.elementAt(2*e);
			if(edgeColor==0)
				return eLeftFront+2;
			return eLeftFront-1;
		}	*/
	}
	
	/**
	 * Perform getRightBack operator: return the index of edge uz, neighbor of edge e=uv
	 */
	public int getRightBack(int e) {
		if(e==e20)
			return e21;
		
		int u=e/3;
		int eColor=e%3;
		
		if(eColor==0) { // red edges
			int bitBack=this.table.getBit2(4*u); // orientation of edge uz
			int bitFront=this.table.getBit3(4*u); // orientation of edge vz

			if(bitBack==0) // edge uz is oriented toward z
				return e+2;
			if(bitFront==1) // edge zv is oriented toward v (and edge uz is oriented toward u)
				return e+4; // e+3+1, because of ordering of red edges
			else
				return this.getRightFront(this.getRightFront(e));
		}
		else if(eColor==1) { // blue edges
			int bitBack=this.table.getBit2(4*u+1); // orientation of edge uz
			int bitFront=this.table.getBit3(4*u+1); // orientation of edge vz
			
			if(bitBack==0) // uz is outgoing from u
				return e-1;
			if(bitFront==1) // zv is oriented toward v (and uz is oriented toward u)
				return this.table.elementAt(4*u+2)+1;
			else
				return this.getRightFront(this.table.elementAt(4*u+2));
		}
		else { // black edges
			int bitBack=this.table.getBit2(4*u+3); // orientation of edge uz
			int bitFront=this.table.getBit3(4*u+3); // orientation of edge vz
			
			if(bitBack==0) // uz is outgoing from u
				return e-1;
			if(bitFront==1) // zv is oriented toward v (and uz is oriented toward u)
				return this.table.elementAt(4*u+3)-2;
			else
				return this.getRightFront(this.table.elementAt(4*u+3));
		}

	}
	
	/**
	 * Return a vertex incident to edge e (e is oriented outgoing from u)
	 */
	public int getSource(int e) {
		return e/3;
	}

	public int getTarget(int e) {
		//System.out.println("target "+e);
		if(e==this.e10) return v0;
		if(e==this.e21) return v1;
		if(e==this.e20) return v0;
		int eColor=e%3;
		//int eLeftFront=this.table.elementAt(2*e);
		int eLeftFront=this.getLeftFront(e);
		int eLeftColor=eLeftFront%3;
		
		if(eLeftColor==eColor)
			return getTarget(eLeftFront);
		else
			return this.getSource(eLeftFront);
	}
	
	/**
	 * Return an edge incident to vertex v
	 */
	public int getEdge(int v) {
		if(v==v0)
			return e10;
		return v*3;
	}
	
	/** 
	 * Check whether edge e is outgoing from vertex v
	 **/
	public boolean isOutgoing(int e, int v) {
		if(e/3==v) return true;
		return false;
	}

	/** 
	 * Return the degree of a given vertex (ccw direction)
	 **/
	/*public int vertexDegreeFast(int v) {
		int result=1;
		int eFirst=this.getEdge(v); // edge of color 0 incident to the vertex
		int pEdge; // edge turning around vertex v
		int ingoing;
		
		if(v!=v0 && v!=v1 && v!=v2) { // testing vertices on the outer face
			pEdge=this.getLeftBack(eFirst);
			ingoing=table.getFirstServiceBit(2*eFirst);
		}
		else
			return vertexDegreeSlow(v);
		
		while(pEdge!=eFirst) {
			if(ingoing==0) {
				ingoing=table.getFirstServiceBit(2*pEdge);
				pEdge=this.getLeftBack(pEdge);
			}
			else {
				ingoing=table.getSecondServiceBit(2*pEdge+1);
				pEdge=this.getRightFront(pEdge);
			}
			
			result++;
		}
		
		return result;
	} */

	/** 
	 * Return the degree of a given vertex (ccw direction)
	 **/
	public int vertexDegreeNormal(int v) {
		int result=1;
		int eFirst=this.getEdge(v); // edge incident to the vertex
		int pEdge; // edge turning around vertex v
		
		if(v!=v0)
			pEdge=this.getLeftBack(eFirst);
		else
			return vertexDegree(v); // LCA: check this case
		
		while(pEdge!=eFirst) {
			//System.out.print("pEdge "+pEdge+": "+this.getSource(pEdge)+"-"+this.getTarget(pEdge));
			if(pEdge/3==v) {
				//System.out.print(" outgoing -> "+this.getRightBack(pEdge));
				pEdge=this.getLeftBack(pEdge);
			}
			else
				pEdge=this.getRightFront(pEdge);
			//System.out.println();
			result++;
		}
		return result;
	}
	
	/** 
	 * Return the degree of a given vertex (cw direction)
	 **/
	public int vertexDegreeMedium(int v) {
		int result=1;
		int eFirst=this.getEdge(v); // edge incident to the vertex
		int pEdge; // edge turning around vertex v
		
		if(v!=v0)
			pEdge=this.getRightBack(eFirst);
		else
			pEdge=this.getLeftFront(eFirst);
		
		while(pEdge!=eFirst) {
			//System.out.print("pEdge "+pEdge+": "+this.getSource(pEdge)+"-"+this.getTarget(pEdge));
			if(pEdge/3==v) {
				//System.out.print(" outgoing -> "+this.getRightBack(pEdge));
				pEdge=this.getRightBack(pEdge);
			}
			else
				pEdge=this.getLeftFront(pEdge);
			//System.out.println();
			result++;
		}
		return result;
	}

	/** 
	 * Return the degree of a given vertex
	 **/
	public int vertexDegree(int v) {
		int result=1;
		int eFirst=this.getEdge(v); // edge incident to the vertex
		int pEdge; // edge turning around vertex v
		
		if(this.isOutgoing(eFirst, v)==true)
			pEdge=this.getRightBack(eFirst);
		else
			pEdge=this.getLeftFront(eFirst);
		
		while(pEdge!=eFirst) {
			if(this.isOutgoing(pEdge, v)==true) {
				pEdge=this.getRightBack(pEdge);
			}
			else
				pEdge=this.getLeftFront(pEdge);
			//System.out.println();
			result++;
		}
		return result;
	}

	/**
	 * Return the normal to a vertex in the mesh (simple float precision)
	 * Slow version, turning in cw direction
	 */
	/* public float[] getNormalFloatFast(int v) {
		//System.out.print(v0+" "+v1+" "+v2);
		//System.out.println("normal "+v);
		if(v==v0 || v==v1 || v==v2)
			return getNormalFloatSlow(v);

		//System.out.println("Vertex normal compact triangle"); // debug
		float[] result=new float[3]; // average of cross product vectors
		float[] prod; // cross product vector
		int degree=1; // vertex degree
		int ingoing; // say whether an edge is outgoing from v
		
		int u, w; // indices of vertices u and v
		float[] pV, pU, pW; // coordinates of vertices v, u and w
		float[] uv=new float[3]; // coordinates of vector V-U
		float[] uw=new float[3]; // coordinates of vector W-U

		int edge=this.getEdge(v); 	// edge incident to vertex v
		int pEdge;					// edge turning around vertex v
		//if(this.getSource(edge)==v) { // the edge is outgoing from v
			pEdge=this.getLeftBack(edge);
			ingoing=table.getFirstServiceBit(2*edge);
			u=this.getTarget(edge);
			v=this.getSource(edge);
		//}
		//else { // the edge is ingoing (having v as target vertex)
		//	pEdge=this.getLeftFront(edge);
		//	v=this.getTarget(edge);
		//	u=this.getSource(edge);
		}
		//if(this.getSource(pEdge)==v) // pEdge is outgoing from v
			w=this.getTarget(pEdge);
		else // pEdge is ingoing (having v as target vertex)
			w=this.getSource(pEdge);

		pV=this.coordF[v]; // coordinates of vertex v
		pU=this.coordF[u]; // coordinates of vertex u
		pW=this.coordF[w]; // coordinates of vertex w
		
		//System.out.println(vectorToString(pV)+" - "+vectorToString(pU)+" - "+vectorToString(pW)); // debug

		// computing vectors V-U and W-U
		uv[0]=pV[0]-pU[0]; uv[1]=pV[1]-pU[1]; uv[2]=pV[2]-pU[2];
		uw[0]=pW[0]-pU[0]; uw[1]=pW[1]-pU[1]; uw[2]=pW[2]-pU[2];	
		result=crossProduct(uv, uw);
		pU=pW;
		uv=uw;
		
		while(pEdge!=edge) {
			//System.out.println("pEdge "+pEdge+" - edge "+edge);
			if(ingoing==0) {
				ingoing=table.getFirstServiceBit(2*pEdge);
				pEdge=this.getLeftBack(pEdge);
				//pEdge=this.getRightBack(pEdge);
			}
			else {
				ingoing=table.getSecondServiceBit(2*pEdge+1);
				pEdge=this.getRightFront(pEdge);
				//pEdge=this.getLeftFront(pEdge);
			}
			if(this.getSource(pEdge)==v) // pEdge is outgoing from v
				w=this.getTarget(pEdge);
			else // pEdge is ingoing (having v as target vertex)
				w=this.getSource(pEdge);
			
			pW=this.coordF[w]; // updating coordinates of vertex w	
			
			//System.out.println(vectorToString(pV)+" - "+vectorToString(pU)+" - "+vectorToString(pW)); // debug
			
			uw[0]=pW[0]-pU[0]; 
			uw[1]=pW[1]-pU[1]; 
			uw[2]=pW[2]-pU[2];
			prod=crossProduct(uv, uw);

			result[0]=result[0]+prod[0];
			result[1]=result[1]+prod[1];
			result[2]=result[2]+prod[2];

			pU=pW;
			uv=uw;
			degree++;
		}	
		result[0]=result[0]/degree;
		result[1]=result[1]/degree;
		result[2]=result[2]/degree;
		
		//System.out.println("degree: "+degree);
		return result;			
	} */
	
	/**
	 * Return the normal to a vertex in the mesh (simple float precision)
	 * Slow version, turning in cw direction
	 */
	public float[] getNormalFloat(int v) {
		float[] result=new float[3]; // average of cross product vectors
		float[] faceNormal; // cross product vector
		int degree=1; // vertex degree
		int u, w; // indices of vertices u and v
		float[] pV, pU, pW; // coordinates of vertices v, u and w

		int edge=this.getEdge(v); 	// edge incident to vertex v
		int pEdge;					// edge turning around vertex v
		if(this.getSource(edge)==v) { // the edge is outgoing from v
			pEdge=this.getRightBack(edge);
			u=this.getTarget(edge);
			v=this.getSource(edge);
		}
		else { // the edge is ingoing (having v as target vertex)
			pEdge=this.getLeftFront(edge);
			v=this.getTarget(edge);
			u=this.getSource(edge);
		}
		if(this.getSource(pEdge)==v) // pEdge is outgoing from v
			w=this.getTarget(pEdge);
		else // pEdge is ingoing (having v as target vertex)
			w=this.getSource(pEdge);

		/*pV=this.coordF[v]; // coordinates of vertex v
		pU=this.coordF[u]; // coordinates of vertex u
		pW=this.coordF[w]; // coordinates of vertex w*/
		pV=this.getPoint(v); // coordinates of vertex v
		pU=this.getPoint(u); // coordinates of vertex u
		pW=this.getPoint(w); // coordinates of vertex w
		
		result=NavigationInterface.computeFaceNormal(pU, pV, pW);
		pU=pW;
		
		while(pEdge!=edge) {
			if(this.getSource(pEdge)==v) {
				pEdge=this.getRightBack(pEdge);
			}
			else {
				pEdge=this.getLeftFront(pEdge);
			}
			if(this.getSource(pEdge)==v) // pEdge is outgoing from v
				w=this.getTarget(pEdge);
			else // pEdge is ingoing (having v as target vertex)
				w=this.getSource(pEdge);
			
			pW=this.getPoint(w); // updating coordinates of vertex w	
			
			faceNormal=NavigationInterface.computeFaceNormal(pU, pV, pW);

			result[0]=result[0]+faceNormal[0];
			result[1]=result[1]+faceNormal[1];
			result[2]=result[2]+faceNormal[2];

			pU=pW;
			degree++;
		}	
		result[0]=result[0]/degree;
		result[1]=result[1]/degree;
		result[2]=result[2]/degree;
		
		return result;			
	}
	
	//--- methods for visualizing the data structure ---
	
	/**
	 * Return a String which represents the reference table encoding the data structure
	 */
	public String toString() {
		String result="Compact Triangle Mesh DS (4 ref per vertex)\n";
		
		for(int i=0;i<this.nVertices;i++) {
			result=result+"v "+i+":";
			for(int j=0;j<this.nReferences;j++) {
				result=result+"\t"+this.table.elementToString(i*this.nReferences+j);
			}
			result=result+"\t"+this.getPoint(i)[0]+","+this.getPoint(i)[1]+"\n";
		}
		
		return result;
	}

	/**
	 * Return the name of the data structure
	 */
	public String name() {
		return "CDS4n";
	}

}
