package compactTriangulations.bench;

import compactTriangulations.CompactTriangleMesh_4n;
import compactTriangulations.CompactTriangleMesh_5n;
import compactTriangulations.CompactTriangleMesh_6n;
import compactTriangulations.CompactTriangleMesh_6nNew;
import arraybasedDS.ArrayBasedHalfedge;
import arraybasedDS.ArrayBasedWingededge;
import arrays.TableWithTwoServiceBits;

//use JVM option -Xms512M or -Xms1024M for increasing heap memory size

/**
 * @author Luca Castelli Aleardi
 * This class provides methods for measuring runtime performances of mesh data structures.
 */
public class EvaluateNavigationalOperations {
	static int repeat=100;
	static long startTime, endTime;
	static double duration;
	//static double[] timings=new double[10]; // experimental runtime performances

	
	
	// ----------------------------------------------------
	// ----------------- Benchmarking ---------------------
	// ----------------------------------------------------
	
	
	
	/**
	 * Evaluating the performance of array based implementation of half edge DS
	 */
	public static double testVertexDegree(ArrayBasedHalfedge ds) {
		System.out.print("Array based Half-edge DS: \t");
				
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex degree (ns): "+duration);
		return duration;
	}
	
	/**
	 * Benchmarking the vertex normal operator, for array based half-edge (for 3D meshes)
	 */
	public static double testVertexNormal(ArrayBasedHalfedge ds) {
		System.out.print("Array based Half-edge: \t");

		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex normal (ns): "+duration);
		return duration;
	}
	
	/**
	 * Evaluating Compact triangle DS (using 6 references per vertex)
	 */
	public static double testVertexDegree(CompactTriangleMesh_6n ds) {
		System.out.print("Compact triangle 6n: \t");
		
		// warming phase
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		
		// start measuring performances
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex degree (ns): "+duration);
		return duration;
	}
	
	/**
	 * Performance of vertex normal (simple float precision) for Compact 6n
	 */
	public static double testVertexNormal(CompactTriangleMesh_6n ds) {
		System.out.print("Compact triangle 6n: \t");
				
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex normal (float precision) (ns): "+duration);
		return duration;
	}

	/**
	 * Evaluating Compact triangle DS (using 5 references per vertex)
	 */
	public static double testVertexDegree(CompactTriangleMesh_5n ds) {
		System.out.print("Compact triangle 5n: \t");
		
		// warming phase
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		
		// start measuring performances
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex degree (ns): "+duration);
		return duration;
	}
	
	/**
	 * Performance of vertex normal (simple float precision) for Compact 5n
	 */
	public static double testVertexNormal(CompactTriangleMesh_5n ds) {
		System.out.print("Compact triangle 5n: \t");
				
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex normal (float precision) (ns): "+duration);
		return duration;
	}

	/**
	 * Evaluating Compact triangle DS (using 4 references per vertex)
	 */
	public static double testVertexDegree(CompactTriangleMesh_4n ds) {
		System.out.print("Compact triangle 4n: \t");
		
		// warming phase
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		
		// start measuring performances
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex degree (ns): "+duration);
		return duration;
	}
	
	/**
	 * Performance of vertex normal (simple float precision) for CDS4n
	 */
	public static double testVertexNormal(CompactTriangleMesh_4n ds) {
		System.out.print("Compact triangle 4n: \t");
				
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex normal (float precision) (ns): "+duration);
		return duration;
	}

	/**
	 * Evaluating Compact triangle DS New(using 6 references per vertex)
	 */
	public static double testVertexDegree(CompactTriangleMesh_6nNew ds) {
		System.out.print("Compact triangle 6n (New): \t");
		
		// warming phase
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++) {
				ds.vertexDegree(i);
				//System.out.println("v"+i+": "+degree);
			}
		}		
		
		// start measuring performances
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex degree (ns): "+duration);
		return duration;
	}

	/**
	 * Performance of vertex normal (simple float precision) for Compact 6n
	 */
	public static double testVertexNormal(CompactTriangleMesh_6nNew ds) {
		System.out.print("Compact triangle 6n (new): \t");
				
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex normal (float precision) (ns): "+duration);
		return duration;
	}

	/**
	 * Performance of 'getTarget' for Compact 6n
	 */
	public static double testGetTarget(CompactTriangleMesh_6nNew ds) {
		System.out.print("Compact triangle 6n: \t");
		System.out.println("Testing 'getTarget'");
				
		for(int j=0;j<1;j++) {
			for(int i=3;i<ds.sizeOfVertices();i++) {
				int color=0;
				int e=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(i, color);
				System.out.print("(v"+i+", "+color+") - ");
				int target=ds.getTarget(e);
				System.out.println("v"+target);
			}
		}		
		startTime=System.nanoTime();
		/*for(int j=0;j<repeat;j++) {
			for(int i=1;i<ds.sizeOfVertices();i++)
				ds.getTarget(i);
		}	*/	
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex normal (float precision) (ns): "+duration);
		return duration;
	}

	/**
	 * Performance of 'getLeftFront' for Compact 6n
	 */
	public static double testGetLeftFront(CompactTriangleMesh_6nNew ds) {
		System.out.print("Compact triangle 6n: \t");
		System.out.println("Testing 'getLeftFront'");
				
		for(int j=0;j<1;j++) {
			for(int i=3;i<ds.sizeOfVertices();i++) {
				int color=2;
				int e=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(i, color);
				System.out.print("(v"+i+", "+color+") ---> ");
				int eLeftFront=ds.getLeftFront(e);
				
				int u=TableWithTwoServiceBits.extractReferenceFromInteger(eLeftFront);
				int colorLeftFront=TableWithTwoServiceBits.extractTwoServiceBitsFromInteger(eLeftFront);
				System.out.println("(v"+u+", "+colorLeftFront+")");
			}
		}		
		startTime=System.nanoTime();
		/*for(int j=0;j<repeat;j++) {
			for(int i=1;i<ds.sizeOfVertices();i++)
				ds.getTarget(i);
		}	*/	
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex normal (float precision) (ns): "+duration);
		return duration;
	}

	/**
	 * Performance of 'getLeftFront' for Compact 6n
	 */
	public static void testGetRightFront(CompactTriangleMesh_6nNew ds) {
		System.out.print("Compact triangle 6n: \t");
		System.out.println("testing 'getRightFront'");
				
		int[] allEdges=getAllEdges(ds);
		
		for(int j=0;j<allEdges.length;j++) {
			int e=allEdges[j];
			System.out.print(ds.printEdge(e)+" ---> ");
			
			int neighbor=ds.getRightFront(e);
				
			System.out.println(ds.printEdge(neighbor));
		}
		
	}

	/**
	 * Performance of 'getLeftFront' for Compact 6n
	 */
	public static void testGetRightBack(CompactTriangleMesh_6nNew ds) {
		System.out.print("Compact triangle 6n: \t");
		System.out.println("testing 'getRightBack'");
				
		int[] allEdges=getAllEdges(ds);
		
		for(int j=0;j<allEdges.length;j++) {
			int e=allEdges[j];
			System.out.print(ds.printEdge(e)+" ---> ");
			
			int neighbor=ds.getRightBack(e);
				
			System.out.println(ds.printEdge(neighbor));
		}
		
	}

	/**
	 * Evaluating array based Winged-edge DS (vertex degree operator)
	 */
	public static double testVertexDegree(ArrayBasedWingededge ds) {
		System.out.print("Array based Winged-edge: \t");
				
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.vertexDegree(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex degree (ns): "+duration);
		return duration;
	}
	
	/**
	 * Performance of vertex normal (simple float precision) for Winged-edge
	 */
	public static double testVertexNormal(ArrayBasedWingededge ds) {
		System.out.print("Array based Winged-edge: \t");
				
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		startTime=System.nanoTime();
		for(int j=0;j<repeat;j++) {
			for(int i=0;i<ds.sizeOfVertices();i++)
				ds.getNormalFloat(i);
		}		
		endTime=System.nanoTime();
		duration=((endTime-startTime)/(double)ds.sizeOfVertices())/repeat;
		System.out.println("vertex normal (float precision) (ns): "+duration);
		return duration;
	}
	
	
	
	// ----------------------------------------------------
	// ----------------- Auxiliary methods ----------------
	// ----------------------------------------------------

	
	
	public static void showMemory() {
		Runtime runtime = Runtime.getRuntime();  
		  
		long maxMemory = runtime.maxMemory();  
		long allocatedMemory = runtime.totalMemory();  
		long freeMemory = runtime.freeMemory();  
		  
		System.out.print("free memory: " + freeMemory / 1024);  
		System.out.print(", allocated: " + allocatedMemory / 1024);  
		System.out.print(", max: " + maxMemory /1024);  
		System.out.println(", total free memory: " + (freeMemory + (maxMemory - allocatedMemory)) / 1024);	
	}
	
	public static String performancesToString(String[] operators, double[] timings) {
		String result="Experimental performances\n";
		for(int i=0;i<operators.length;i++)
			result=result+operators[i]+" (ns): "+timings[i]+"\n";
		return result;
	}
	
	/**
	 * Return the collection of all 'valid' edges, encoded as pairs: (u, color)
	 */
	public static int[] getAllEdges(CompactTriangleMesh_6nNew ds) {
		int nEdges=ds.sizeOfVertices()*3 -6;
		//System.out.println("vertices "+ds.sizeOfVertices());
		int[] result=new int[ds.sizeOfVertices()*3 -6]; // there are exactly 3n-6 edges
		System.out.print("Generating all edges ...");
		
		int count=0;
		for(int i=0;i<ds.sizeOfVertices();i++) {
			for(int color=0;color<3;color++) {
				//System.out.println(i+"-"+color+": "+count);
				int e=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(i, color);
				
				if(ds.isEdge(e)) {
					result[count]=e;
					count++;
				}
			}
		}
		
		System.out.println("done ("+result.length+")");
		return result;
	}

}
