package compactTriangulations.building;

import java.util.*;

import Jcg.geometry.Point_;
import Jcg.polyhedron.*;
import Jcg.schnyderwoods.*;

/**
 * @author Luca Castelli Aleardi
 * This class provides functions for creating compact array based data structures for triangle meshes
 * 
 * Remark: for an Halfedge e, e.index is the original number (index) in the mesh
 * getEdgeIndex(e) is the (new) number of the edge, according to the Schnyder wood labeling of edges
 * In particular, for an edge of color c (0,1 or 2), incident to vertex i (0..n-1),
 * the number of e is : 3*i+c
 */
public abstract class CompactMeshBuilder {
	
	PlanarTriSchnyderWood schnyder; // stores the orientation/coloration of edges corresponding to a Schnyder Wood
	Polyhedron_3<Point_> polyhedron; // input mesh
	byte[] coloration; // coloration of half-edges
	boolean[] orientation; // orientation of half-edges
	
	public CompactMeshBuilder(PlanarTriSchnyderWood cut) {
		this.schnyder=cut;
		this.polyhedron=(Polyhedron_3<Point_>)cut.polyhedron;
		
		this.orientation=this.schnyder.getEdgeOrientation();
		this.coloration=this.schnyder.getEdgeColoration();
	}
		
	/**
	 * Return the number (index) of an edge, given its color (0, 1 or 2), and the index of the incident vertex
	 * Recall that the edge must be oriented outgoing from the vertex 
	 */
	protected int getEdgeIndex(int color, int vertexIndex) {
		return vertexIndex*3+color;
	}
	
	/**
	 * Return the number (index) of an (half)edge: edges are ordered according to
	 * the vertex numbering and their color.
	 * Recall that edges must be colored according to the Schnyder wood definition
	 */
	protected int getEdgeIndex(Halfedge e) {
		int eColor=coloration[e.index];
		int vIndex; // number of the incident vertex (edge e must be oriented outgoing from v)
		if(orientation[e.index]==true)
			vIndex=e.getOpposite().getVertex().index;
		else vIndex=e.getVertex().index;
		return getEdgeIndex(eColor, vIndex);
	}

	/**
	 * Return the number (index) of an (half)edge: edges are ordered according to
	 * the vertex numbering and their color.
	 * Recall that edges must be colored according to the Schnyder wood definition
	 * In this version, vertices are permuted according to BFS ordering on T0
	 */
	protected int getEdgeIndex(Halfedge e, int[] permutation) {
		int eColor=coloration[e.index];
		int vIndex; // number of the incident vertex (edge e must be oriented outgoing from v)
		if(orientation[e.index]==true)
			vIndex=permutation[e.getOpposite().getVertex().index];
		else vIndex=permutation[e.getVertex().index];
		return getEdgeIndex(eColor, vIndex);
	}

	/**
	 * Return a permutation of input vertices according to a BFS traversal of tree T0
	 */
	protected int[] getBFSPermutation() {
		int[] permutation=new int[this.polyhedron.sizeOfVertices()];
		int counter=0;
		LinkedList<Halfedge> edges=new LinkedList<Halfedge>(); // a queue for traversing in BFS order the tree T0
		
		// add all halfedges incident to v0 to the queue
		edges.add(this.schnyder.rootEdge.getOpposite()); // add halfedge e10 (having v0 as target vertex)
		Halfedge pEdge=this.schnyder.rootEdge.getPrev();
		while(pEdge!=this.schnyder.rootEdge.getOpposite()) {
			edges.add(pEdge);
			pEdge=pEdge.getOpposite().getPrev();
		}
		
		permutation[counter]=this.schnyder.v0.index; // the first vertex is the root node of edge T0
		counter++;
		
		// perform the BFS traversal of tree T0
		while(edges.isEmpty()==false) {
			Vertex v=edges.poll().getOpposite().getVertex(); // retrieve the source vertex of the first edge in the queue
			permutation[counter]=v.index;
			counter++;
			
			Halfedge first=this.schnyder.getFirstIncomingRedEdge(v); // retrieve the first incoming red edge incident to v
			if(first!=null) { // if there is at least one incoming red edge
				edges.add(first);
				Halfedge next=this.schnyder.getNextIncomingRedEdge(first);
				while(next!=null) {
					edges.add(next);
					next=this.schnyder.getNextIncomingRedEdge(next);
				}
			}
			
		}
		//System.out.println("counter "+counter);
		return permutation;
	}
	
	/**
	 * Return the inverse of a permutation
	 */
	protected int[] inversePermutation(int[] permutation) {
		int[] inverse=new int[permutation.length];
		
		for(int i=0;i<permutation.length;i++)
			inverse[permutation[i]]=i;
		
		return inverse;
	}

	/**
	 * Check the edge orientations and vertex labels
	 */
	public boolean checkOrientation(){
		System.out.print("Checking edge orientations and vertex labels...");

		// checking vertex labels
		int i=0;
		for(Vertex<Point_> v: this.polyhedron.vertices) {
			if(v.index<0 || v.index>=polyhedron.sizeOfVertices())
				throw new Error("wrong vertex labels: "+v.index);
			i++;
		}

		// checking half-edge labels and edge orientation
		i=0;
		for(Halfedge<Point_> e:this.polyhedron.halfedges) {
			if(e.index<0)
				throw new Error("wrong edge labels: "+e.index);
			if(orientation[e.index]==orientation[e.getOpposite().index])
				throw new Error("wrong half-edge orientation: "+e.index);
			i++;
		}
		System.out.println("ok");
		
		return true;
	}

	/**
	 * Print the orientation and coloring of all vertices
	 */
	public String printOrientation(){
		String result="Printing edge colorings and orientations\n";

		// checking half-edge labels and edge orientation
		for(Halfedge<Point_> e:this.polyhedron.halfedges) {
			if(orientation[e.index]==true) {
				int eSource=e.getOpposite().getVertex().index;
				result=result+"(v"+eSource+" ,v"+e.getVertex().index+") ";
				result=result+"color "+coloration[e.index]+"\n";
			}

		}
		
		return result;
	}

	/**
	 * Return c+1 (mod 3)
	 */
	public int nextColor(int color) {
		return (color+1)%3; // basic implementation (involving modulo 3 division)
	}

}
