package compactTriangulations.building;

import compactTriangulations.CompactTriangleMesh_5n;

import Jcg.geometry.Point_;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Vertex;
import Jcg.schnyderwoods.PlanarTriSchnyderWood;

/**
 * @author Luca Castelli Aleardi
 * This class provides functions for creating compact array based data structures for triangle meshes
 * 
 * Remark: for an Halfedge e, e.index is the original number (index) in the mesh
 * getEdgeIndex(e) is the (new) number of the edge, according to the Schnyder wood labeling of edges
 * In particular, for an edge of color c (0,1 or 2), incident to vertex i (0..n-1),
 * the number of e is : 3*i+c
 */
public class CompactMeshBuilder_5n extends CompactMeshBuilder {

	public CompactMeshBuilder_5n(PlanarTriSchnyderWood cut) {
		super(cut);
		this.checkOrientation();
	}
	
	/**
	 * Construct the compact array based data structure requiring 5 reference per vertex
	 * Half-edges must be oriented. 
	 * Half-edges must have labels between 0..2E-1
	 * Vertices must have labels between 0..N-1
	 */
	public CompactTriangleMesh_5n buildFromPolyhedron(){
		System.out.print("Creating a compact triangle mesh DS (5n rpv) from a Polyhedron [CDS5n]...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

    	CompactTriangleMesh_5n result;
		int nV=this.polyhedron.sizeOfVertices();
		
		result=new CompactTriangleMesh_5n(nV);
		
		// checking vertex labels
		int i=0;
		for(Vertex<Point_> v: this.polyhedron.vertices) {
			if(v.index<0 || v.index>=polyhedron.sizeOfVertices())
				throw new Error("wrong vertex labels: "+v.index);
			i++;
		}

		// checking half-edge labels and edge orientation
		i=0;
		for(Halfedge<Point_> e:this.polyhedron.halfedges) {
			if(e.index<0)
				throw new Error("wrong edge labels: "+e.index);
			if(orientation[e.index]==orientation[e.getOpposite().index])
				throw new Error("wrong half-edge orientation: "+e.index);
			i++;
		}
		
		// initializing (inner) edges
		for(Halfedge<Point_> e: this.polyhedron.halfedges) {
			setEdge(result,e);
		}
		
		// initializing vertex coordinates
		result.coordF=polyhedron.coordF;

		/*for(Vertex<Point_> v: this.polyhedron.vertices) {
			Point_ p=v.getPoint();
			result.setPoint(v.index, 	(float)p.getCartesian(0).doubleValue(), 
										(float)p.getCartesian(1).doubleValue(), 
										(float)p.getCartesian(2).doubleValue());
		}*/
		
		// initializing vertices on the root face
		result.v0=this.schnyder.v0.index;
		result.v1=this.schnyder.v1.index;
		result.v2=this.schnyder.v2.index;
		
		// initializing edges on the root face
		result.e10=result.v1*3;
		result.e20=result.v2*3;
		result.e21=result.v2*3+1;
		
		// setting neighboring relations for the 3 edges on the root face
		// the leftFront neighbor of edge e10 is a dummy (not existing) edge of color 1 incident to v0	
		result.setLeft(result.e10, result.v0+1); // leftFront neighbor of e10
		result.setLeft(result.e21, result.e10); // the leftFront neighbor of e21 is e10
		result.setRight(result.e20, result.e10); // the rightFront neighbor of e20 is e10
		
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
		
		return result;
	}
	
	/**
	 * Initialize the information concerning edge e
	 */
	public void setEdge(CompactTriangleMesh_5n ds, Halfedge e) {
		if(orientation[e.index]==false) return;
			
		// setting references to LeftFront and RightFront incident edges
		int eIndex=getEdgeIndex(e);
		int eColor=coloration[e.index];
		int leftBackColor=coloration[e.getPrev().index];
		int rightBackColor=coloration[e.getOpposite().getNext().index];
		int leftFrontColor=coloration[e.getNext().index];
		int rightFrontColor=coloration[e.getOpposite().getPrev().index];
		int blueEdgeColoration=coloration[e.getOpposite().getPrev().getOpposite().getPrev().index];
		
		if(eColor==0) { // setting red edges
			// setting eLeftFront
			ds.setLeft(eIndex, this.getEdgeIndex(e.getNext()));
			
			// setting rightFront (depends on the orientation of neighboring edges)
			if(rightFrontColor!=2 || blueEdgeColoration!=1)
				ds.setRight(eIndex, this.getEdgeIndex(e.getOpposite().getPrev()));
			else
				ds.setRight(eIndex, this.getEdgeIndex(e.getOpposite().getPrev().getOpposite().getPrev()));
			
			// setting service bits (depending on orientation/coloration of LeftBack and RightBack edges)
			if(leftBackColor==(eColor+2)%3)
				ds.setLeftBit0(eIndex);
			if(rightBackColor==(eColor+1)%3)
				ds.setRightBit0(eIndex);
			
			if(leftFrontColor==eColor)
				ds.setLeftBit1(eIndex);
			if(rightFrontColor==eColor)
				ds.setRightBit1(eIndex);
		}
		else if(eColor==1) { // setting blue edges
			// setting eLeftFront, eRightFront
			ds.setLeft(eIndex, this.getEdgeIndex(e.getNext()));
			ds.setRight(eIndex, this.getEdgeIndex(e.getOpposite().getPrev()));
			
			// setting service bits (depending on orientation/coloration of LeftBack and RightBack edges)
			if(leftBackColor==(eColor+2)%3)
				ds.setLeftBit0(eIndex);
			if(rightBackColor==(eColor+1)%3)
				ds.setRightBit0(eIndex);
			
			if(leftFrontColor==eColor)
				ds.setLeftBit1(eIndex);
			if(rightFrontColor==eColor)
				ds.setRightBit1(eIndex);
		}
		else { // setting black edges
			// setting eLeftFront or eRightFront (depending on the orientation of neighboring edges)
			if(rightBackColor==1)
				ds.setLeft(eIndex, this.getEdgeIndex(e.getNext()));
			else
				ds.setRight(eIndex, this.getEdgeIndex(e.getOpposite().getPrev()));
			
			// setting service bits (depending on orientation/coloration of LeftBack and RightBack edges)
			if(leftBackColor==(eColor+2)%3)
				ds.setLeftBit0(eIndex);
			if(rightBackColor==(eColor+1)%3)
				ds.setRightBit0(eIndex);
			
			if(leftFrontColor==eColor)
				ds.setLeftBit1(eIndex);
			if(rightFrontColor==eColor)
				ds.setRightBit1(eIndex);
		}

	}
	
}
