package compactTriangulations.building;

import compactTriangulations.CompactTriangleMesh_6n;

import Jcg.geometry.Point_;
import Jcg.polyhedron.*;
import Jcg.schnyderwoods.PlanarTriSchnyderWood;

/**
 * @author Luca Castelli Aleardi
 * This class provides functions for creating compact array based data structures for triangle meshes
 * 
 * Remark: for an Halfedge e, e.index is the original number (index) in the mesh
 * getEdgeIndex(e) is the (new) number of the edge, according to the Schnyder wood labeling of edges
 * In particular, for an edge of color c (0,1 or 2), incident to vertex i (0..n-1),
 * the number of e is : 3*i+c
 */
public class CompactMeshBuilder_6n extends CompactMeshBuilder {

	public CompactMeshBuilder_6n(PlanarTriSchnyderWood cut) {
		super(cut);
		this.checkOrientation();
	}
	
	/**
	 * Construct an compact array based data structure requiring 6 reference per vertex
	 * Half-edges must be oriented. 
	 * Half-edges must have labels between 0..2E-1
	 * Vertices must have labels between 0..N-1
	 */
	public CompactTriangleMesh_6n buildFromPolyhedron(){
		System.out.print("Creating a compact triangle mesh DS (6n rpv) from a Polyhedron [CDS6n]...");
		int nV=this.polyhedron.sizeOfVertices();
		
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

    	CompactTriangleMesh_6n result=new CompactTriangleMesh_6n(nV);
				
		// initializing edges
		for(Halfedge<Point_> e: this.polyhedron.halfedges) {
			if(orientation[e.index]==true) {
				// setting references to LeftFront and RightFront incident edges
				int eIndex=this.getEdgeIndex(e);
				int eColor=coloration[e.index];
				result.setLeft(eIndex, this.getEdgeIndex(e.getNext()));
				result.setRight(eIndex, this.getEdgeIndex(e.getOpposite().getPrev()));
				
				// setting service bits (depending on orientation/coloration of LeftBack and RightBack edges)
				int leftBackColor=coloration[e.getPrev().index];
				int rightBackColor=coloration[e.getOpposite().getNext().index];
				if(leftBackColor==(eColor+2)%3)
					result.setLefServiceBit(eIndex);
				if(rightBackColor==(eColor+1)%3)
					result.setRightServiceBit(eIndex);
				
				int leftFrontColor=coloration[e.getNext().index];
				int rightFrontColor=coloration[e.getOpposite().getPrev().index];
			}
		}
		
		// initializing vertex coordinates
		result.coordF=polyhedron.coordF;
		
		/*for(Vertex<Point_> v: this.polyhedron.vertices) { // old version (not efficient)
			Point_ p=v.getPoint();
			result.setPoint(v.index, this.getX, 
										(float)p.getCartesian(1).doubleValue(), 
										(float)p.getCartesian(2).doubleValue());
		}*/
		
		// initializing vertices on the root faces
		result.v0=this.schnyder.v0.index;
		result.v1=this.schnyder.v1.index;
		result.v2=this.schnyder.v2.index;
		
		// initializing edges on the root faces
		result.e10=result.v1*3;
		result.e20=result.v2*3;
		result.e21=result.v2*3+1;
		
		// setting neighboring relations for the 3 edges on the root face
		// the leftFront neighbor of edge e10 is a dummy (not existing) edge of color 1 incident to v0	
		result.setLeft(result.e10, result.v0+1); // leftFront neighbor of e10
		result.setLeft(result.e21, result.e10); // the leftFront neighbor of e21 is e10
		result.setRight(result.e20, result.e10); // the rightFront neighbor of e20 is e10
		
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
		
		return result;
	}	
	
}
