package compactTriangulations.building;

import Jcg.geometry.Point_;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Vertex;
import Jcg.schnyderwoods.PlanarTriSchnyderWood;
import arrays.TableWithTwoServiceBits;
import compactTriangulations.CompactTriangleMesh_6nNew;

/**
 * @author Luca Castelli Aleardi
 * This class provides functions for creating compact array based data structures for triangle meshes
 * 
 * Remark: for an Halfedge e, e.index is the original number (index) in the mesh
 * getEdgeIndex(e) is the (new) number of the edge, according to the Schnyder wood labeling of edges
 * In particular, for an edge of color c (0,1 or 2), incident to vertex i (0..n-1),
 * the number of e is : 3*i+c
 */
public class CompactMeshBuilder_6nNew extends CompactMeshBuilder {


	public CompactMeshBuilder_6nNew(PlanarTriSchnyderWood cut) {
		super(cut);
		this.checkOrientation();
		//System.out.println(this.printOrientation());
	}
	
	/**
	 * Construct an compact array based data structure requiring 6 reference per vertex
	 * Half-edges must be oriented. 
	 * Half-edges must have labels between 0..2E-1
	 * Vertices must have labels between 0..N-1
	 */
	public CompactTriangleMesh_6nNew buildFromPolyhedron(){
		System.out.print("Creating a compact triangle mesh DS (6n rpv) from a Polyhedron [CDS6n slow]...");
		int nV=this.polyhedron.sizeOfVertices();
		
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

    	CompactTriangleMesh_6nNew result=new CompactTriangleMesh_6nNew(nV);
    	
		// initializing vertices on the root face
		result.v0=this.schnyder.v0.index;
		result.v1=this.schnyder.v1.index;
		result.v2=this.schnyder.v2.index;
		
		// set edges on outer (root) face
		result.e10=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(result.v1, 0); // e10, red edge outgoing from v1
		result.e20=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(result.v2, 0); // e20, red edge outgoing from v2
		result.e21=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(result.v2,1); // e21, blue edge outgoing from v2
				
		// initializing edges
		for(Halfedge<Point_> e: this.polyhedron.halfedges) {
			if(orientation[e.index]==true) {
				// setting references to LeftFront and RightFront incident edges
				int eColor=coloration[e.index];
				int eSource=e.getOpposite().getVertex().index; // source of 'e'
				
				// store leftFront edge
				Halfedge eLeftFront=e.getNext(); // edge (v, w)
				int colorLeftFront=coloration[eLeftFront.index];
				int sourceLeftFront; // source vertex of leftFront edge
				
				if(orientation[eLeftFront.index]==true) // the edge (v, w) is oriented toward w
					sourceLeftFront=e.getVertex().index;
				else // (w, v) is oriented toward v
					sourceLeftFront=eLeftFront.getVertex().index;
				
				result.setU(eSource, sourceLeftFront, eColor, 0); // set left reference (left='0')
				
				if(eColor==colorLeftFront)
					result.setT(eSource, eColor, 0); // set service bit
				// --- end setting leftFront edge ---
				
				// store rightFront edge
				Halfedge eRightFront=e.getOpposite().getPrev();
				int colorRightFront=coloration[eRightFront.index];
				int sourceRightFront; // source vertex of rightFront edge
				
				if(orientation[eRightFront.index]==true) // the edge (z, v) is oriented toward v
					sourceRightFront=eRightFront.getOpposite().getVertex().index;
				else // (v, z) is oriented toward z
					sourceRightFront=eRightFront.getVertex().index;
				
				result.setU(eSource, sourceRightFront, eColor, 1); // set right reference (right='1')
				
				if(eColor==colorRightFront)
					result.setT(eSource, eColor, 1);
				// --- end setting rightFront edge ---				
			}
		}
		
		// initializing incident incoming edges (service bits in the three arrays S)
		for(Halfedge<Point_> e: this.polyhedron.halfedges) {
			if(orientation[e.index]==true) {
				int eColor=coloration[e.index];
				int eSource=e.getOpposite().getVertex().index; // source of 'e'
				
				// set booleans in the three arrays S (encoding the existence of incoming edges)
				// solution: given an outgoing edge of color c, check whether its rightBack edge is incoming (of color c+1)
				// remark: edge (2, red) has not incoming blue edges at its right, by definition
				if(eSource!=result.v2 || eColor!=0) { // handling all edges, except (v2, v0)
					Halfedge eRightBack=e.getOpposite().getNext();
					if(orientation[eRightBack.index]==false) {// the edge is incoming (oriented toward v)
						result.setS(eSource, nextColor(eColor));
					}
				}
			}
		}

		// initializing vertex coordinates
		result.coordF=polyhedron.coordF;
		
		/*for(Vertex<Point_> v: this.polyhedron.vertices) {
			float x=polyhedron.getX(v.index);
			float y=polyhedron.getY(v.index);
			float z=polyhedron.getZ(v.index);
			result.setPoint(v.index, x, y, z);
		}*/
		
		// setting neighboring relations for the 3 edges on the root face
		// the leftFront neighbor of edge e10 is a dummy (not existing) edge of color 1 incident to v0	
		/*result.setLeft(result.e10, result.v0+1); // leftFront neighbor of e10
		result.setLeft(result.e21, result.e10); // the leftFront neighbor of e21 is e10
		result.setRight(result.e20, result.e10); // the rightFront neighbor of e20 is e10*/
		
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
		
		return result;
	}	

}
