package compactTriangulations.encoding;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

import compactTriangulations.test.MemoryBenchmark;

/**
 * This class provides methods for reading and writing binary files, and for converting Integers/Floats from/to byte[] <p>
 * The bytes in the binary encoding are read/written with a linear scan
 * <p>
 * Remark: this class requires at least Java 1.7
 * 
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2016)
 * 
 */
public class BinaryEncoding {

	/** Array of bytes storing the binary encoding*/
	protected byte[] encoding;

	/** Counter of bytes: used to perform a linear scan of the binary encoding (read a byte at time) */
	protected int byteCounter=0;

	/** Counter of bits: used to perform a linear scan of the binary encoding (read a bit at time) */
	protected int bitCounter=0;

	/**
	 * Warning: this method is not supported

	 */
	public BinaryEncoding() {
	    this.encoding=null;
	}

	/**
	 * Create a binary encoding of a given size
	 * 
	 * @param b  the size of the encoding (number of bytes)
	 */
	public BinaryEncoding(int b) {
	    this.encoding=new byte[b];
	}

	/**
	 * Read the binary encoding from file
	 * 
	 * @param filename name of the input file
	 */
	public BinaryEncoding(String filename) throws IOException {
		long memory0=MemoryBenchmark.getUsedMemory(); // used memory before constructing the triangle mesh

		System.out.print("Reading binary file...");
	    Path path = Paths.get(filename);
	    this.encoding=Files.readAllBytes(path);
        long memory1=MemoryBenchmark.getUsedMemory(); // used memory before constructing the triangle mesh
	    System.out.println("done");
        MemoryBenchmark.printMemory(memory1-memory0, "[loadede binary compressed file in main memory]");
	}
	
	/**
	 * Write the whole encoding to a binary file
	 * 
	 * @param filename name of the output file
	 */
	public void writeToFile(String filename) throws IOException {
	    Path path = Paths.get(filename);
	    Files.write(path, this.encoding); // creates or overwrites
	  }
	
	/**
	 * Read the next 4 bytes (a memory word, on 32 bits)
	 * 
	 * @return an array of 4 bytes storing a memory word (on 32 bits)
	 */
	public byte[] getNextWord() {
		byte[] result=new byte[4];
		result[0]=this.encoding[byteCounter];
		result[1]=this.encoding[byteCounter+1];
		result[2]=this.encoding[byteCounter+2];
		result[3]=this.encoding[byteCounter+3];
		
		byteCounter=byteCounter+4;
		return result;
	}

	/**
	 * Append to the encoding the next 4 bytes (a memory word, on 32 bits)
	 * 
	 * @param word  an array of 4 bytes storing a memory word (on 32 bits)
	 */
	public void appendNextWord(byte[] word) {
		this.encoding[byteCounter]=word[0];
		this.encoding[byteCounter+1]=word[1];
		this.encoding[byteCounter+2]=word[2];
		this.encoding[byteCounter+3]=word[3];
		
		byteCounter=byteCounter+4; // increase the number of stored memory words
	}

	/**
	 * Read the next 4 bytes and return the corresponding integer value
	 * 
	 * @return a 32 bits integer
	 */
	public int getNextInteger() {
		byte[] nextWord=this.getNextWord();
		return fromByteArray(nextWord);
	}

	/**
	 * Read the next 4 bytes and return the corresponding integer value
	 * 
	 * @return a 32 bits integer
	 */
	public int getNextBit() {
		byte b=this.encoding[this.byteCounter];
		int result=readBitInByte(b, this.bitCounter);
		
		this.bitCounter++;
		if(this.bitCounter%8==0) {
			this.bitCounter=0; // reset the bit counter: scan from the beginning the byte (next time)
			this.byteCounter++; // scan next byte
		}
		return result;
	}

	/**
	 * Read the i-th in the byte
	 * 
	 * @return an the bit at position 'i'
	 */
	public int readBitInByte(byte b, int i) {
		return (int)(b>>i);
	}

	/**
	 * Set the i-th in the byte to '1'
	 */
	public byte setBitInByte(byte b, int i) {
		throw new Error("to be completed");
	}

	/**
	 * Append to the encoding an integer on 32 bits (4 bytes)
	 * 
	 * @param v  an integer integer (4 bytes) to append to the encoding
	 */
	public void appendNextInteger(int v) {
		byte[] word=toByteArray(v);
		
		this.encoding[byteCounter]=word[0];
		this.encoding[byteCounter+1]=word[1];
		this.encoding[byteCounter+2]=word[2];
		this.encoding[byteCounter+3]=word[3];
		
		byteCounter=byteCounter+4; // increase the number of stored memory words
	}

	/**
	 * Read the next 4 bytes and return the corresponding float value
	 * 
	 * @return a 32 bits float
	 */
	public float getNextFloat() {
		byte[] nextWord=this.getNextWord();
		
		float result=Float.intBitsToFloat(fromByteArray(nextWord)); // convert to integer and then to float
		return result;
	}

	/**
	 * Append to the encoding a float number on 32 bits (4 bytes)
	 * 
	 * @param x  a float value (4 bytes) to append to the encoding
	 */
	public void appendNextFloat(float x) {
		byte[] word=toByteArray(Float.floatToIntBits(x)); // convert the float to integer, and then to a byte array
		
		this.encoding[byteCounter]=word[0];
		this.encoding[byteCounter+1]=word[1];
		this.encoding[byteCounter+2]=word[2];
		this.encoding[byteCounter+3]=word[3];
		
		byteCounter=byteCounter+4; // increase the number of stored memory words
	}

	/**
	 * Decode an array of four bytes to the corresponding integer value (on 32 bits)
	 * 
	 * @param bytes an array of four bytes encoding a 32 bits integer
	 * @return an integer, encoded as 4 bytes
	 */
	public static int fromByteArray(byte[] bytes) {
	     return bytes[0] << 24 | (bytes[1] & 0xFF) << 16 | (bytes[2] & 0xFF) << 8 | (bytes[3] & 0xFF);
	}

	/**
	 * Encode an integer value (on 32 bits) as four bytes
	 * 
	 * @param value an integer value to encode
	 * @return bytes an array of four bytes encoding a 32 bits integer
	 */
	public static byte[] toByteArray(int value) {
	    return new byte[] { 
	        (byte)(value >> 24),
	        (byte)(value >> 16),
	        (byte)(value >> 8),
	        (byte)value };
	}

	/**
	 * Convert an array of integers to an array of bytes
	 * 
	 * @param bits  an array of integers storing the encoding
	 * @return an array of bytes storing the same encoding
	 */
	private static byte[] fromIntArrayToByteArray(int[] bits) {
		int n=bits.length; // number of words in the array of integers
		int b=n*4; // number of bytes words in the resulting array
		byte[] result=new byte[b];
		
		int count=0; // byte counter
		for(int i=0;i<n;i++) {
			int intValue=bits[i]; // the i-th integer value
			byte[] bytes=toByteArray(intValue); // the four bytes corresponding to the i-th integer value
			
			// set the four bytes corresponding to the i-th integer
			result[count]=bytes[0];
			result[count+1]=bytes[1];
			result[count+2]=bytes[2];
			result[count+3]=bytes[3];
			
			count=count+4;
		}
		
		return result;
	}

	/**
	 * Return the byte counter: the number of bytes already read/written during the linear scan of the encoding
	 * 
	 * @return the number of bytes already read/written during the linear scan of the encoding
	 */
	public int getByteCounter() {
		return this.byteCounter;
	}

}
