package compactTriangulations.encoding;

import java.util.HashMap;
import java.util.HashSet;

import Jcg.geometry.Point_3;
import Jcg.mesh.HashMapEdges;
import Jcg.mesh.SharedVertexRepresentation;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.HalfedgePair;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import compactTriangulations.test.MemoryBenchmark;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2016)
 * 
 * This class provides methods converting and reading OFF files stored with binary encoding <p>
 * <p>
 * Remark: this class requires at least Java 1.7
 * 
 */
public class BinaryOFF {

	/**
	 * Read a triangle mesh in standard OFF format (text file) and convert it into a binary format <p>
	 * <p>
	 * Warning: it works only for planar triangle meshes (without boundaries)
	 */
	public static void fromOFFToBinaryOFF(String input, String output) {
		SharedVertexRepresentation m=new SharedVertexRepresentation(input); // read coordinates and faces from OFF
		
		System.out.println("Converting OFF to binary");
		int n=m.sizeVertices;
		int f=2*n-4; // number of faces (in a planar triangulation, without holes)
		int size=4+(3*n*4)+(3*f*4); // story 4 bytes for each coordinates; stores 3*4 bytes for each face
		
		if(m.faces.length!=f) {
			System.err.println("Error: the input mesh is not homeomorphic to a sphere");
			System.exit(0);
		}
		
		BinaryEncoding binary=new BinaryEncoding(size);
		
		// header
		binary.appendNextInteger(n); // write the header: 'n'
		System.out.println("\tEncoding file header: "+n);
		
		System.out.print("\tEncoding vertex coordinates...");
		for (Point_3 p : m.points) { // write vertex coordinates (4 bytes for each coordinate)
			binary.appendNextFloat((float)p.getX().floatValue());
			binary.appendNextFloat((float)p.getY().floatValue());
			binary.appendNextFloat((float)p.getZ().floatValue());
			//System.out.println(p);
		}
		System.out.println("done");
		
		// write the indices of vertices in each face
		System.out.print("\tEncoding faces...");
		for (int i=0; i<f; i++) {
			for (int j=0; j<3; j++) { // the faces are assumed to be triangles
				binary.appendNextInteger(m.faces[i][j]);
			}
		}
		System.out.println("done");
		
		try {
			binary.writeToFile(output); // write the encoding to an output binary file
		} catch (Exception e) {}
		System.out.println("\tBinary encoding saved to file ("+binary.getByteCounter()+" bytes): "+output);
	}

	/**
	 * Read a triangle mesh in binary OFF format and construct the corresponding shared vertex representation <p>
	 * <p>
	 * Warning: it works only for planar triangle meshes (without boundaries)<p>
	 * <p>
	 * The input format is the following<p>
	 * <p>
	 * n<p>
	 * x1 y1 z1<p>
	 * x2 y2 z2<p>
	 * ... ... ...<p>
	 * v11 v12 v13<p>
	 * v21 v22 v23<p>
	 * v31 v32 v33<p>
	 * v41 v42 v43<p>
	 * 
	 * <p>
	 * All numbers (integers and floats) are encoding on 32 bits (four bytes each)
	 * spaces between numbers and other characters are not encoded
	 */
	public static SharedVertexRepresentation fromBinaryOFFToSharedRepresentation(String input) {
		System.out.println("Reading a 'shared vertex representation' from binary OFF file: "+input);
		BinaryEncoding binary=null;
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		try{
			binary=new BinaryEncoding(input);
		} catch(Exception e) {}
		
		if(binary==null)
			throw new Error("Error: wrong input file: "+input);
		
		int n=binary.getNextInteger();
		int f=2*n-4;
		int he=2*(3*n-6);
		int[][] faces=new int[f][3];
		int[] faceDegrees=new int[f];
		Point_3[] points=new Point_3[n];
		System.out.println("\tReading header n="+n);
		
		System.out.print("\tReading coordinates...");
		for(int i=0;i<n;i++) {
			double x=binary.getNextFloat();
			double y=binary.getNextFloat();
			double z=binary.getNextFloat();
			
			points[i]=new Point_3(x, y, z);
			//System.out.println(points[i]);
		}
		System.out.println("done");
		
		System.out.print("\tReading faces...");
		for(int i=0;i<f;i++) {
			faceDegrees[i]=3; // faces are triangle (by assumption)
			
			int u=binary.getNextInteger(); // first vertex in the face
			int v=binary.getNextInteger(); // second vertex
			int w=binary.getNextInteger(); // third vertex
			faces[i][0]=u;
			faces[i][1]=v;
			faces[i][2]=w;
		}
		System.out.println("done");
		
        endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        System.out.println("Mesh loaded from binary OFF format ("+duration+" seconds)");

		return new SharedVertexRepresentation(n, f, he, faces, faceDegrees, points);
	}

	/**
	 * Read a triangle mesh in binary OFF format and directly construct the corresponding Half-edge representation.
	 * It does not requires to store in main memory the 'shared vertex representation': the half-edge data structures
	 * is allocated on the fly.
	 * <p>
	 * Warning: it works only for planar triangle meshes (without boundaries): all face degrees are assumed to be 3. <p>
	 * <p>
	 * The input format is the following<p>
	 * <p>
	 * n<p>
	 * x1 y1 z1<p>
	 * x2 y2 z2<p>
	 * ... ... ...<p>
	 * v11 v12 v13<p>
	 * v21 v22 v23<p>
	 * v31 v32 v33<p>
	 * v41 v42 v43<p>
	 * 
	 * <p>
	 * All numbers (integers and floats) are encoding on 32 bits in the input file (using four bytes each).
	 * Spaces between numbers and other characters are not encoded
	 */
	public static Polyhedron_3 fromBinaryOFFToPolyhedron(String input) {
		long memory0=MemoryBenchmark.getUsedMemory(); // used memory before constructing the triangle mesh
		
		System.out.println("Building a pointer-based half-edge data structure (for genus 0 triangle meshes) from a binary OFF file: "+input);
		BinaryEncoding binary=null;
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		try{
			binary=new BinaryEncoding(input);
		} catch(Exception e) {}
		
		if(binary==null)
			throw new Error("Error: wrong input file: "+input);
		
		int n=binary.getNextInteger();
		int f=2*n-4; // the input mesh is assumed to be a closed planar triangulation (genus 0)
		int he=2*(3*n-6);
		System.out.println("\tReading header n="+n);
		    	
    	Polyhedron_3<Point_3> result=new Polyhedron_3<Point_3>(n,he,f); // initialize the resulting triangle mesh of size 'n'
        		
		System.out.print("\tReading coordinates...");
		for(int i=0;i<n;i++) {
        	Vertex<Point_3> v=new Vertex<Point_3>();
        	v.index=i;
            result.vertices.add(v); // adding a vertex
            
            // store its geometric coordinates
            int index=i*3;
        	result.coordF[index]=binary.getNextFloat();
        	result.coordF[index+1]=binary.getNextFloat();
        	result.coordF[index+2]=binary.getNextFloat();
			//System.out.println(points[i]);
		}
		System.out.println("done");
		
		System.out.print("\tReading faces...");
		int[] faceVertices=new int[3]; // store for each face index of the 3 incident vertices

		//HashMap<Halfedge<Point_3>,Integer> insertedHalfedges=new HashMap(he/2); // allows to efficient retrieving an inserted half-edge
		int sizeHashMap=(int)(he/6*(3.5)); // tune the size of the hashmap to get better runtime performances or memory storage
		HashMapEdges map=new HashMapEdges(he);

		Halfedge<Point_3> e0=null;
		Halfedge<Point_3>[] _edges=new Halfedge[3];
		
		int edgeCounter=0;
		//int countEdgesHashMap=0;
		//int countOppositeEdgesHashMap=0;
		Halfedge e;
        for(int i=0;i<f;i++) {
        	faceVertices[0]=binary.getNextInteger(); // first vertex in the face
        	faceVertices[1]=binary.getNextInteger(); // second vertex
        	faceVertices[2]=binary.getNextInteger(); // third vertex
            
            // creating and storing the 'd' half-edges in a face
            for(int j=0;j<3;j++) { // d=3
            	e=new Halfedge<Point_3>();
            	result.halfedges.add(e);
            	e.index=edgeCounter;
            	edgeCounter++;
            	_edges[j]=e; // storing the edges of the face
            	
            	// setting incident (target) vertex
            	e.setVertex(result.vertices.get(faceVertices[(j+1)%3]));
            	// half-edges must be incident to the current face
            }
            
            // setting incidence relations between half-edges in the same face
            for(int j=0;j<3;j++) { // d=3
            	int indNext=(j+1)%3; // index of next half-edge // d=3
            	int indPrev; // index of previous half-edge
            	if(j>0) indPrev=j-1;
            	else indPrev=3-1; // d=3
            	
            	_edges[j].setNext(_edges[indNext]);
            }
            
            // setting opposite half-edge
            for(int j=0;j<3;j++) { // d=3
            	e0=_edges[j];
            	
            	//int u=e0.vertex.index;
            	//int v=e0.next.next.vertex.index;
            	//int eIndex=e0.index;
            	
            	int insertedEdge=map.put(e0.vertex.index, e0.next.next.vertex.index, e0.index);
            	if(insertedEdge!=-1) { // the half-edge was already stored
            		int indexOpposite=insertedEdge;
            		Halfedge<Point_3> eOpposite=result.halfedges.get(indexOpposite);
            		_edges[j].setOpposite(eOpposite);
            		eOpposite.setOpposite(_edges[j]);
            		
            		//System.out.println("retrieving opposite halfedge of ("+u+", "+v+", "+eIndex+") in hashmap: e"+indexOpposite);
            		//countOppositeEdgesHashMap++;
            	}
            	//else {
            		//System.out.println("inserting halfedge ("+u+", "+v+", "+eIndex+") in hashmap");
            		//countEdgesHashMap++;
            	//}
            	
            	// implementation using the standard Java HashMap
            	/*if(insertedHalfedges.containsKey(e0)==false) {
            		insertedHalfedges.put(e0, e0.index);
            		System.out.println("inserting halfedge ("+u+", "+v+", "+eIndex+") in hashmap");
            		//System.out.println("inserted edge "+e0.index+": "+e0.next.next.vertex.index+" - "+e0.vertex.index);
            		countEdgesHashMap++;
            	}
            	else {
            		//System.out.println("\t edge already inserted");
            		int indexOpposite=insertedHalfedges.get(e0);
            		Halfedge<Point_3> eOpposite=result.halfedges.get(indexOpposite);
            		_edges[j].setOpposite(eOpposite);
            		eOpposite.setOpposite(_edges[j]);
            		countOppositeEdgesHashMap++;
            	}*/
            }
            
        }
        
        // setting the vertex incident to a half-edge
		for(Halfedge pedge: result.halfedges) {
        	Vertex<Point_3> pvertex=pedge.vertex;
        	pvertex.setEdge(pedge);
        }

		System.out.println("done");
		
		//System.out.println("Edges inserted in hash map: "+countEdgesHashMap+", opposite edges: "+countOppositeEdgesHashMap);
		
		// perform memory and runtime benchmarks
        endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        System.out.println("Mesh loaded from binary OFF format ("+duration+" seconds)");
        long memory1=MemoryBenchmark.getUsedMemory(); // used memory before constructing the triangle mesh
        //MemoryBenchmark.printMemory(memory1-memory0, "[pointer-based half-edge DS]");
        MemoryBenchmark.printMemory(memory1, "[total used memory]");
        
        System.out.print("Running garbage collector...");
        runGC();
        System.out.println("done");

		return result;
	}
	
	public static void runGC() {
		//allocateMemory();
		//System.gc(); // call garbage collector
		//System.gc();
		//System.gc();
		//System.gc();
		Runtime r = Runtime.getRuntime();
		r.gc();
	}
	
	public static void allocateMemory() {
		int n=10000000;
		
		Point_3[] points=new Point_3[n];
		for(int i=0;i<n;i++)
			points[i]=new Point_3();
			
	}
	
	public static void vertexDistribution(Polyhedron_3<Point_3> mesh) {
		System.out.println("Degree distribution:");
		
		int n=mesh.vertices.size();
		
		int[] degrees=new int[n];
		for(Vertex v: mesh.vertices) {
			degrees[mesh.vertexDegree(v)]++;
		}
		
		for(int i=3;i<Math.min(n, 20);i++)
			System.out.println("\tdeg "+i+": "+((double)degrees[i]/n)*100+" ("+degrees[i]+" vertices)");
	}

}
