package compactTriangulations.encoding;

import Jcg.geometry.Point_;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.schnyderwoods.PlanarTriSchnyderWood;
import Jcg.util.BitSet32;
import arrays.ArrayBasedStack;
import arrays.ColumnArrayBasedStack;
import arrays.StackInterface;
import arrays.TableWithTwoServiceBits;
import compactTriangulations.CompactTriangleMesh_6nNew;
import compactTriangulations.test.MemoryBenchmark;
import tc.TC;

/**
 * @author Luca Castelli Aleardi
 * 
 * This class implements the decoding algorithm which constructs, from a binary encoding scheme, a Compact Triangle Mesh data structure <p>
 * 
 * Remark: recall that <p>
 * -) all vertices are ordered according to a Depth first traversal of the red tree (rooted at v0) <p>
 * -) the two outer vertices v0 and v1 are ordered as usual: their numbers are '0' and '1'
 * -) v2 is the last vertex: it has index n-1 <p>
 * -) the inner face incident to (v0, v1) is (0, 1, 2)
 * -) the outer face is: (1, 0, n-1)
 * 
 */
public class MeshDecoder {
	private final static int RED=0, BLUE=1, BLACK=2; // edge colors
	private final static int LEFT=0, RIGHT=1; // to distinguish left and right front/back edges
	
	private int N; // global counter for vertices
	private StackInterface redStack, blackStack; // the two red and black stacks (storing indices of red and black edges during the construction phase)
	private int nVertices; // number of vertices
	private CompactTriangleMesh_6nNew cds=null; // the compact data structure
	private StringEncoderTwoWords encoding; // 4n bits encoding of the triangulation (two words)	
	public float[] coordF; // geometric coordinates of vertices
	
	public boolean noAdditionalMemory, binaryFormat;

	// for debugging
	private static int countRecursiveCalls=0;
	//private int[] blueInDegree; // store the number incoming blue edges (for each vertex v1, ..., v_{n-2})
	
	/**
	 * Initialize the decoder: the triangulation is assumed to be encoded by a pair of red and black words
	 * stored in an input file in text format
	 */
	public MeshDecoder(String inputFile, boolean noAdditionalMemory, boolean binaryFormat) {
		this.binaryFormat=binaryFormat;
		this.noAdditionalMemory=noAdditionalMemory;
		
		if(binaryFormat==false) {
			this.encoding=this.readEncodingFromFile(inputFile); // read the red and black words from a text file
		}
		else { // the input file is encoded in binary format
			this.encoding=this.readEncodingFromBinaryFile(inputFile); // read the red and black words from a binary file
		}
		this.nVertices=encoding.n;
		
		/*this.cds=new CompactTriangleMesh_6nNew(this.nVertices); // initialize the data structure

		if(noAdditionalMemory==true) { // do not use additional arrays in the compact data structure: 'blue columns' are used for storing the stacks
			System.out.print(" (no additional memory)");
			this.redStack=new ColumnArrayBasedStack(this.cds.table.elements, 2, 6, this.nVertices); // use column 3 of the table
			this.blackStack=new ColumnArrayBasedStack(this.cds.table.elements, 3, 6, this.nVertices); // use column 4 of the table
		}
		else { // use two additional arrays for storing the red and black edges
			System.out.print(" (using additional memory: two stacks)");
			this.redStack=new ArrayBasedStack(this.nVertices);
			this.blackStack=new ArrayBasedStack(this.nVertices);
		}*/
		
		//this.blueInDegree=new int[this.nVertices]; // for checking correctness (blue tree construction)
		//this.encoding.check();
		
		this.encoding.resetGlobalCounters(); // reset global counters, for restarting the scan of the word from the beginning
	}

	/**
	 * Initialize the decoder: the triangulation is assumed to be encoded by a pair of red and black words
	 * stored in an input file in text format
	 */
	/*public MeshDecoder(StringEncoderTwoWords encoding, boolean noAdditionalMemory) {
		this.nVertices=encoding.n;
		this.cds=new CompactTriangleMesh_6nNew(this.nVertices); // initialize the data structure
		System.out.print("Mesh decoder initialization: "+this.nVertices+" vertices");

		if(noAdditionalMemory==true) { // do not use additional arrays in the compact data structure: 'blue columns' are used for storing the stacks
			System.out.print(" (no additional memory)");
			this.redStack=new ColumnArrayBasedStack(this.cds.table.elements, 2, 6, this.nVertices); // use column 3 of the table
			this.blackStack=new ColumnArrayBasedStack(this.cds.table.elements, 3, 6, this.nVertices); // use column 4 of the table
		}
		else { // use two additional arrays for storing the red and black edges
			System.out.print(" (using additional memory: two stacks)");
			this.redStack=new ArrayBasedStack(this.nVertices);
			this.blackStack=new ArrayBasedStack(this.nVertices);
		}
		System.out.println();
		
		this.encoding=encoding;
		
		//this.blueInDegree=new int[this.nVertices]; // for checking correctness (blue tree construction)
		//this.encoding.check();
		
		this.encoding.resetGlobalCounters(); // reset global counters, for restarting the scan of the word from the beginning
	}*/

	/**
	 * Read the vertex coordinates and the red and black words from a text file
	 */
	public StringEncoderTwoWords readEncodingFromFile(String inputFile) {
		StringEncoderTwoWords result;
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		System.out.print("Reading geometry from file...");
		TC.lectureDansFichier(inputFile); // start reading data from file
		int n=Integer.parseInt(TC.lireLigne()); // number of vertices (file header)

		// geometric coordinates	
		this.coordF=new float[n*3]; // for storing geometric coordinates
		for(int i=0;i<n;i++) { 
			String coordinates=TC.lireLigne();
			String[] tabCoordinates=TC.motsDeChaine(coordinates);
			if(tabCoordinates.length<3)
				throw new Error("Error: wrong number of geometric coordinates");
			
			float x=Float.parseFloat(tabCoordinates[0]);
			float y=Float.parseFloat(tabCoordinates[1]);
			float z=Float.parseFloat(tabCoordinates[2]);
			
			int index=i*3;
			this.coordF[index]=x;
			this.coordF[index+1]=y;
			this.coordF[index+2]=z;
		}

		endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
		
		// start reading connectivity
		startTime=System.nanoTime();
		System.out.print("Reading connectivity from file...");
		result=new StringEncoderTwoWords(n); // initialize the encoding (for a triangulation with n vertices)
		
		String word=TC.lireLigne(); // red word
		for(int i=0;i<word.length();i++) { // scan the red word in the input file
			if(word.charAt(i)=='(')
				result.addRedChar(1);
			else if(word.charAt(i)==')')
				result.addRedChar(0);
			else
				throw new Error("Error: wrong symbol in encoding red word");
		}
		
		word=TC.lireLigne(); // black word
		for(int i=0;i<word.length();i++) { // scan the black word in the input file
			if(word.charAt(i)=='1')
				result.addBlackChar(1);
			else if(word.charAt(i)=='0')
				result.addBlackChar(0);
			else
				throw new Error("Error: wrong symbol in encoding red word");
		}
		
    	endTime=System.nanoTime();
        duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
        
		TC.lectureEntreeStandard();
		result.resetGlobalCounters();
		return result;
	}

	/**
	 * Read the vertex coordinates and the red and black words from a binary file
	 * <p>
	 * Warning: it works only for 3d triangle meshes (without boundaries)<p>
	 * <p>
	 * The input format is the following<p>
	 * <p>
	 * n<p>
	 * x1 y1 z1<p>
	 * x2 y2 z2<p>
	 * ... ... ...<p>
	 * 010011... // the red word<p>
	 * 000100... // the black word<p>
	 * <p>
	 * All numbers (integers and floats) are encoding on 32 bits (four bytes each)
	 * spaces between numbers and other characters are not encoded
	 */
	public StringEncoderTwoWords readEncodingFromBinaryFile(String inputFile) {
		System.out.println("Reading a compressed triangle mesh from binary file: "+inputFile);
		long startTime, endTime; // for evaluating time performances

		startTime=System.nanoTime();
		StringEncoderTwoWords result; // the encoding based on the red and black words
		BinaryEncoding binary=null; // the compressed representation stored in the binary input file
		
		try{
			binary=new BinaryEncoding(inputFile);
		} catch(Exception e) {}
		
		if(binary==null)
			throw new Error("Error: wrong input file: "+inputFile);
		
		int n=binary.getNextInteger(); // read the header (the number of vertices in the mesh)
		System.out.println("\tReading the header n="+n);

		System.out.print("\tReading vertex coordinates from file...");
		this.coordF=new float[n*3]; // for storing geometric coordinates
		for(int i=0;i<n;i++) {
			
			int index=i*3;
			this.coordF[index]=binary.getNextFloat();
			this.coordF[index+1]=binary.getNextFloat();
			this.coordF[index+2]=binary.getNextFloat();
		}
		endTime=System.nanoTime();
		double durationGeometry=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+durationGeometry+" seconds)");

		System.out.print("\tReading connectivity from file...");
		startTime=System.nanoTime();
		int redBits=2*(n-1); // number of bits in the red code
		int blackBits=2*n-4; // number of bits in the black code
		int redWords=((redBits/32)+1); // number of memory words (32 bits) in the red code
		int blackWords=((blackBits/32)+1); // number of memory words (32 bits) in the black code
		
		int[] redIntArray=new int[redWords];
		byte[] word;
		for(int i=0;i<redWords;i++) {
			word=binary.getNextWord(); // read four bytes from the binary encoding
			redIntArray[i]=BinaryEncoding.fromByteArray(word); // convert four bytes to an integer value
		}
		int[] blackIntArray=new int[blackWords];
		for(int i=0;i<blackWords;i++) {
			word=binary.getNextWord(); // read four bytes from the binary encoding
			blackIntArray[i]=BinaryEncoding.fromByteArray(word); // convert four bytes to an integer value
		}
		
		BitSet32 redCode, blackCode; // the two encodings of the red and black words
		redCode=new BitSet32(redIntArray, redBits);
		blackCode=new BitSet32(blackIntArray, blackBits);
		result=new StringEncoderTwoWords(redCode, blackCode, n); // initialize the encoding (for a triangulation with n vertices)
		result.resetGlobalCounters();
		
		endTime=System.nanoTime();
		double durationConnectivity=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+durationConnectivity+" seconds)");
		System.out.println("\tTotal duration (reading binary file): "+(durationGeometry+durationConnectivity)+" seconds)");
		
		return result;
	}

	/**
	 * Decode and construct the compact representation (using 6n references)
	 */
	public CompactTriangleMesh_6nNew decode() {
		System.out.print("Start decoding");
		long startTime=System.nanoTime(), endTime; // for evaluating time performances

		this.nVertices=encoding.n;
		this.cds=new CompactTriangleMesh_6nNew(this.nVertices); // initialize the data structure
		
		if(noAdditionalMemory==true) { // do not use additional arrays in the compact data structure: 'blue columns' are used for storing the stacks
			System.out.println(" (no use of additional memory)");
			this.redStack=new ColumnArrayBasedStack(this.cds.table.elements, 2, 6, this.nVertices); // use column 3 of the table
			this.blackStack=new ColumnArrayBasedStack(this.cds.table.elements, 3, 6, this.nVertices); // use column 4 of the table
		}
		else { // use two additional arrays for storing the red and black edges
			System.out.println(" (using additional memory: two stacks)");
			this.redStack=new ArrayBasedStack(this.nVertices);
			this.blackStack=new ArrayBasedStack(this.nVertices);
		}

		// initializing vertices on the root face: (0, 1, n-1)
		this.cds.v0=0;
		this.cds.v1=1;
		this.cds.v2=this.nVertices-1;
		
		// set edges on outer (root) face
		this.cds.e10=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(this.cds.v1, 0); // e10, red edge outgoing from v1
		this.cds.e20=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(this.cds.v2, 0); // e20, red edge outgoing from v2
		this.cds.e21=TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(this.cds.v2, 1); // e21, blue edge outgoing from v2
		
		this.N=1; // global counter for vertices
		this.encoding.readNextRedChar(); // the first '(' is already read
		this.encoding.readNextRedChar(); // the first ')' is already read
		this.redStack.push(1);// push the vertex in the stack
		
		System.out.print("First phase (constructing red and black trees)...");
		this.constructRedTree();
		this.setRedEdgesOuterFace(); // set the edges (v1, v0)) and (v2, v0)
		System.out.println("ok");
		
		// reset the two stacks: this is required when the stacks are store in Table T
		this.redStack.reset();
		this.blackStack.reset();
		
		this.constructBlueTree();
		this.setBlueEdgesOuterFace();
		
		this.cds.coordF=this.coordF; // set the geometric coordinates
		
		// evaluate runtime and memory costs
		endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("Decoding and construction done ("+duration+" s)");
		long initUsed=MemoryBenchmark.getUsedMemory();
		long initFree=MemoryBenchmark.getFreeMemory();
		MemoryBenchmark.printMemory(initUsed, "[total used memory]");
		MemoryBenchmark.printMemory(initFree, "[total free memory]");

		// chech vertex degrees in the blue tree (only for debug)
		//this.checkBlueInDegrees();

		return this.cds;
	}
	
	/**
	 * Set the red edges, defining the tree T_0
	 * 
	 * @param u  index of a vertex
	 * @param degree  the number of incoming black edges at vertex u
	 */
	private void constructRedTree() {
		int u=this.redStack.top(); // construct the tree rooted at vertex u
		if(u<0 || u>=this.nVertices)
			throw new Error("Error: wrong vertex index "+u); // u is the index of a vertex and must be in [0..n-1]
		
		int blackInDegree=this.encoding.readNextBlackBlock(); // number of incoming black edges at vertex u
		
		this.constructIncomingBlackEdges(u, blackInDegree);
		
		int redChar=this.encoding.readNextRedChar(); // red parenthesis
		
		if(redChar==0) { // no more descendant edges in the red tree
			// nothing to do
		}
		else if(redChar==1){ // there are incoming red edges (oriented toward 'u')
			this.cds.setS(u, RED);
			this.N++;
			int v=N;
			int w=N; // first red child of vertex 'u' 
			
			if(v<0 || v>=this.nVertices)
				throw new Error("Error: wrong vertex index "+v); // u is the index of a vertex and must be in [0..n-1]

			this.cds.setU(w, u, RED, LEFT); // set the LeftFront edge of (w, u), which is of color blue
			
			while(v==w) { // visit all red sub-trees rooted at 'u'
				this.redStack.push(v);
				//System.out.println("v="+v);
				this.constructRedTree();// recursive call: construction of the sub-tree rooted at 'v'
				v=this.redStack.pop();
				w=v;
				u=this.redStack.top();
				
				if(u<0 || u>=this.nVertices)
					throw new Error("Error: wrong vertex index "+u); // u is the index of a vertex and must be in [0..n-1]
				if(v<0 || v>=this.nVertices)
					throw new Error("Error: wrong vertex index "+v); // v is the index of a vertex and must be in [0..n-1]
				
				int nextChar2=this.encoding.readNextRedChar();
				if(nextChar2==1) {
					this.N++;
					w=N;
					
					// set the LeftFront edge of (w, u)
					this.cds.setU(w, v, RED, LEFT);
					this.cds.setT(w, RED, LEFT);
					
					// set the RightFront edge of (v, u)
					this.cds.setU(v, w, RED, RIGHT);
					this.cds.setT(v, RED, RIGHT);
					v=w;
				}
				else
					w=-1; // to exit the loop
			}
			this.cds.setU(v, u, RED, RIGHT); // set the RightFront edge of the last red child of u (which must be Black)
		}
		
		this.blackStack.push(u);
	}

	/**
	 * Set the black edges incoming at vertex u
	 * 
	 * @param u  index of a vertex
	 * @param degree  the number of incoming black edges at vertex u
	 */
	private void constructIncomingBlackEdges(int u, int degree) {
		if(degree>0) { // there are some black incoming edges
			int x=this.blackStack.pop(); // source of the first incoming black edge (x, u)
			
			// set the leftFront edge of (x, u)
			this.cds.setU(x, u, BLACK, LEFT); 
			this.cds.setS(u, BLACK);
			
			degree--;
			while(degree>0) { // turn around vertex 'u' d-1 times, in ccw order
				int y=x;
				degree--;
				x=this.blackStack.pop(); // next black edge around 'u', after (y, u)
				
				// set the leftFront edge of (x, u)
				this.cds.setU(x, y, BLACK, LEFT); 
				this.cds.setT(x, BLACK, LEFT);
				
				// set the rightFront edge of (y, u)
				this.cds.setU(y, x, BLACK, RIGHT);
				this.cds.setT(y, BLACK, RIGHT);
			}
			// set the rightFront edge of the last black incoming edge
			this.cds.setU(x, u, BLACK, RIGHT);
		}
	}
	
	/**
	 * Set the blue edges
	 */
	private void constructBlueTree() {
		for(int x=1;x<this.nVertices-1;x++) { // iterate over all vertices, except vertex v0 and v2 (which do not have incident blue edges)
			//System.out.println("Blue tree - x"+x);
			
			if(cds.getU(x, RED, RIGHT)==cds.getU(x, BLACK, LEFT) && x!=1 && (this.cds.getT(x, RED, RIGHT)==0 || this.cds.getT(x, BLACK, LEFT)==0)) { // x has no blue child: recall that v1 has always blue children
				//System.out.println("\t no blue child");
				// nothing to do
			}
			else {
				this.cds.setS(x, BLUE); // vertex x has at least one blue child
				
				// set the rightmost blue child of x
				int u; // (x, u) denotes the rightmost blue child of x
				if(this.cds.getT(x, RED, RIGHT)==0) { // the rightFront edge of (x, z) is black
					int z=this.cds.getU(x, RED, RIGHT);
					u=this.cds.getU(z, BLACK, RIGHT);
				}
				else { // the rightFront edge of (x, z) is red
					u=this.cds.getU(x, RED, RIGHT);
				}
				cds.setU(u, x, BLUE, RIGHT);
				// --- end setting (u, x) ---
				
				int v=u;
				int count=1; // for debugging
				while(u==v) { // visit the blue children of vertex x in cw order
					//System.out.println("x"+x+"\t v"+v+"\t u"+u+"\t child "+count);					
					if(this.cds.getS(u, RED)==1) { // the edge (v, u) is red (incoming at u)
						v=u+1; // since v is the first child of u
					}
					else { // (u, v) is black, outgoing from u
						v=this.cds.getU(u, BLACK, RIGHT);
					}
					//System.out.println("\t\t v -> "+v);
					
					this.cds.setU(u, v, BLUE, LEFT); // set the leftFront edge of (u, x)
					
					if(x==1 && v==this.nVertices-1) { // the last child of v1 is reached
						this.cds.setU(v, u, BLUE, RIGHT); // set the rightFront edge of (v, x)
						this.cds.setT(u, BLUE, LEFT); // the left front edge of (u, x) is blue
						this.cds.setT(v, BLUE, RIGHT); // the right front edge of (v, x) is blue
						count++;
						v=-1; // to exit the loop
					}
					else if(this.isLastBlueChild(u, x)==false) { // (u, x) is not the last blue child of x (children are traversed in cw order)
						this.cds.setU(v, u, BLUE, RIGHT); // set the rightFront edge of (v, x)
						this.cds.setT(u, BLUE, LEFT); // the left front edge of (u, x) is blue
						this.cds.setT(v, BLUE, RIGHT); // the right front edge of (v, x) is blue
						u=v; // turn around x in cw order
						count++;
					}
					else
						v=-1; // to exit the loop
					
				}
				//this.blueInDegree[x]=count; // store, for each vertex, the number of incoming blue edges (for checking correctness)
			}
		}
	}

	/**
	 * Check whether the edge (u, x) is the last blue child of x (children are traversed in cw order, from right to left)
	 */
	private boolean isLastBlueChild(int u, int x) {
		if(x==1 && u==this.nVertices-1) // x is the vertex v1: so its last blue child must be (v2, v1)
			return true;
		
		// x is not the vertex v1
		if(this.cds.getS(u, LEFT)==0) { // case 1: no incoming red edges at u
			int uRightFront=this.cds.getU(u, BLACK, RIGHT);
			
			if(uRightFront==x) { // case 1a: (x, v) and (u, v) are black, oriented toward v
				return true;
			}
			else { // case 1b: (v, x) is blue oriented toward x
				return false;
			}
		}
		
		else { // case 2: (v, u) is red, oriented toward u
			if(this.cds.getU(x, BLACK, LEFT)!=u+1)
				return false;
			else {
				if(this.cds.getU(u+1, BLACK, RIGHT)==x)
					return false;
				return true;
			}
		}
		
	}

	/**
	 * Set the red edges on the root (outer) face: (v1, v0) and (v2, v0) <p>
	 * 
	 * Remark: recall that v2 has index n-1, and the inner face incident to (v0, v1) is (0, 1, 2)
	 */
	private void setRedEdgesOuterFace() {
		// set the RightFront edge of (v1, v0)
		this.cds.setU(1, 2, RED, RIGHT); 
		this.cds.setT(1, RED, RIGHT);
		this.cds.setT(2, RED, LEFT);

		// set the LeftFront edge of (v1, v0)
		this.cds.setU(1, this.nVertices-1, RED, LEFT); // set the LeftFront edge of (v1, v0)
		this.cds.setT(1, RED, LEFT);
		
		// set the RightFront edge of (v2, v0)
		this.cds.setU(this.nVertices-1, 1, RED, RIGHT); 
		this.cds.setT(this.nVertices-1, RED, RIGHT);
		
		this.cds.setS(0, RED); // vertex v0 has some incoming red edges
		this.cds.setSFalse(1, RED); // there are no red edges going toward v1
	}

	/**
	 * Set the red edges on the root (outer) face: (v1, v0) and (v2, v0) <p>
	 * 
	 * Remark: recall that v2 has index n-1, and the inner face incident to (v0, v1) is (0, 1, 2)
	 */
	private void setBlueEdgesOuterFace() {
		// set the LeftFront edge of (v2, v1)
		this.cds.setU(this.nVertices-1, 1, BLUE, LEFT);
	}

	/**
	 * Check whether the number of incoming blue edges is correct
	 */
	private boolean checkBlueInDegrees() {
		System.out.print("\t Checking construction blue tree (incoming blue degrees)...");
		int degree=0;
		for(int i=0;i<this.nVertices;i++) {
			//System.out.println("v"+i+": "+this.blueInDegree[i]);
			//degree=degree+this.blueInDegree[i];
		}
		
		if(degree==(this.nVertices-1)-1) { // the blue tree must have (n-1) nodes, and thus (n-1)-1 edges
			System.out.println("ok");
			return true;
		}
		else {
			System.out.println("error +("+degree+")");
			return false;
		}
		
	}

}
