package compactTriangulations.encoding;

import java.util.LinkedList;

import Jcg.geometry.Point_;
import Jcg.geometry.Point_3;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import Jcg.schnyderwoods.PlanarTriSchnyderWood;
import tc.TC;

/**
 * @author Luca Castelli Aleardi
 * 
 * This class performs the encoding of a planar triangulation which is endowed with a Schnyder wood orientation
 * <p>
 * The triangulation is encoded with a bit string of size 4n: <p>
 * -) the red tree is traversed in a Depth First manner: its edges are encoded with a balanced parenthesis word, of size 2(n-1) (the vertex v0 is not encoded) <p>
 * -) the remaining bits represent the in-degree of black edges: number of black edges oriented toward a given vertex.
 * 
 */
public class MeshEncoder {
	/** stores the orientation/coloration of edges corresponding to a Schnyder Wood */
	private PlanarTriSchnyderWood schnyder;
	/** the input mesh: a planar triangulation */
	private Polyhedron_3<Point_> polyhedron;
	private byte[] coloration; // coloration of half-edges
	private boolean[] orientation; // orientation of half-edges
	
	/** encoding of the mesh (based on two binary strings): it provides methods for scanning (for reading and writing) the string encoding the mesh */
	public StringEncoderTwoWords encoding;
	
	/**
	 * Initialize the encoder
	 * 
	 * @param cut  the planar triangulation endowed with a Schnyder wood
	 */
	public MeshEncoder(PlanarTriSchnyderWood cut) {
		this.schnyder=cut;
		this.polyhedron=(Polyhedron_3<Point_>)cut.polyhedron;
		
		this.orientation=this.schnyder.getEdgeOrientation();
		this.coloration=this.schnyder.getEdgeColoration();
		
		this.checkOrientation();
		this.encoding=new StringEncoderTwoWords(this.polyhedron.sizeOfVertices());
	}

	/**
	 * Perform the encoding of the triangulation into a binary string
	 */
	public void encode() {
		System.out.print("Encoding a planar triangulation...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

		Halfedge e10=this.schnyder.rootEdge.getOpposite(); // half-edge oriented toward v0
		Halfedge e20=e10.getNext().getOpposite(); // half-edge oriented toward v0 
		Halfedge eStart=this.schnyder.rootEdge.getNext().getNext(); // first sibling of the edge (v1, v0): the second child in T0, oriented toward the root vertex v0
		
		if(this.getEdgeColor(e10)!=0 || this.getEdgeColor(e10)!=0 || this.getEdgeColor(eStart)!=0) // these three edges must be red
			throw new Error("Error: wrong edge color");
		if(this.getEdgeOrientation(e10)==false || this.getEdgeOrientation(e10)==false || this.getEdgeOrientation(eStart)==false) // these three edges must be red
			throw new Error("Error: wrong edge orientation");
		
		// encode the two red parenthesis of vertex v1
		this.encoding.addRedChar(1); 
		this.encoding.addRedChar(0);
		this.encoding.addBlackBlock(0); // vertex v1 has no incoming black edges
		
		Halfedge e=eStart;
		while(e!=e20) { // e20 is the last half-edge to be visited
			//System.out.println(this.printEdge(e));
			
			if(this.getEdgeColor(e)==0 && this.getEdgeOrientation(e)==true) { // Case 1 (outgoing red edge): visiting a new vertex 'v' in the red three
				this.encoding.addRedChar(1); // symbol '('
				e=e.getPrev().getOpposite(); // turn around vertex 'v' in ccw order ('v' is the source of the edge)
			}
			else if(this.getEdgeColor(e)==2 && this.getEdgeOrientation(e)==false) { // Case 2 (incoming black edges)
				while(this.getEdgeColor(e)==2) { // count the number of black incoming edges around 'v'
					this.encoding.addBlackChar(1);
					e=e.getPrev().getOpposite();
				}
			}
			else if(this.getEdgeColor(e)==1 && this.getEdgeOrientation(e)==true) { // Case 3: outgoing blue edge
				this.encoding.addBlackChar(0); // end of the block black incoming edges
				
				if(this.getEdgeColor(e.getPrev())==0) // vertex 'v' has descendants in the red tree
					e=e.getPrev();
				else
					e=e.getPrev().getOpposite(); // visit next outgoing black edge
			}
			else if(this.getEdgeColor(e)==2 && this.getEdgeOrientation(e)==true) { // Case 4: outgoing black edge
				this.encoding.addRedChar(0); // closing parenthesis ')'
				if(this.getEdgeColor(e.getPrev())==0) // move to its red ancestor
					e=e.getPrev();
				else
					e=e.getPrev().getOpposite(); // visit the first incoming blue edge
			}
			else if(this.getEdgeColor(e)==1 && this.getEdgeOrientation(e)==false) { // Case 5: incoming blue edge (nothing to encode)
				if(this.getEdgeColor(e.getPrev().getOpposite())==0)
					e=e.getPrev();
				else
					e=e.getPrev().getOpposite();
			}
			else if(this.getEdgeColor(e)==0 && this.getEdgeOrientation(e)==false) { // Case 6: incoming red edge (leaving the node, moving to its red parent)
				if(this.getEdgeColor(e.getPrev())==0)
					e=e.getPrev(); // move to its next red sibling
				else
					e=e.getPrev().getOpposite(); // moving the black outgoing edge (of the ancestor node)
			}
		}
		
		// deal with last vertex v2
		e=e.getPrev().getOpposite(); // first black incoming edge (oriented toward v2)
		while(this.getEdgeColor(e)==2) { // visit all black incoming edges oriented toward v2
			this.encoding.addBlackChar(1);
			e=e.getPrev().getOpposite(); // visit next edge incident to v2
		}
		this.encoding.addBlackChar(0); // add last bit to the black string
		this.encoding.addRedChar(1); // add the last '(' parenthesis corresponding to e20
		this.encoding.addRedChar(0); // add the last ')' parenthesis corresponding to e20
		
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
	}

	/**
	 * Return the color of an half-edge
	 */
	protected int getEdgeColor(Halfedge e) {
		return this.coloration[e.index];
	}

	/**
	 * Return the color of an half-edge
	 */
	protected boolean getEdgeOrientation(Halfedge e) {
		return this.orientation[e.index];
	}

	/**
	 * Check the edge orientations and vertex labels
	 */
	public boolean checkOrientation(){
		System.out.print("Checking edge orientations and vertex labels...");

		// checking vertex labels
		int i=0;
		for(Vertex<Point_> v: this.polyhedron.vertices) {
			if(v.index<0 || v.index>=polyhedron.sizeOfVertices())
				throw new Error("wrong vertex labels: "+v.index);
			i++;
		}

		// checking half-edge labels and edge orientation
		i=0;
		for(Halfedge<Point_> e:this.polyhedron.halfedges) {
			if(e.index<0)
				throw new Error("wrong edge labels: "+e.index);
			if(orientation[e.index]==orientation[e.getOpposite().index])
				throw new Error("wrong half-edge orientation: "+e.index);
			i++;
		}
		System.out.println("ok");
		
		return true;
	}

	/**
	 * Print the orientation and coloring of all vertices
	 */
	public String printOrientation(){
		String result="Printing edge colorings and orientations\n";

		// checking half-edge labels and edge orientation
		for(Halfedge<Point_> e:this.polyhedron.halfedges) {
			if(orientation[e.index]==true) {
				int eSource=e.getOpposite().getVertex().index;
				result=result+"(v"+eSource+" ,v"+e.getVertex().index+") ";
				result=result+"color "+coloration[e.index]+"\n";
			}

		}
		
		return result;
	}

	/**
	 * Write the encoding of the triangulation to a file (in text format)
	 */
	public void writeToFile(String output){
		System.out.println("Writing compressed format to text file ("+output+")");
		TC.ecritureDansNouveauFichier(output);
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int n=this.polyhedron.sizeOfVertices();
		TC.println(n);
		System.out.print("\tencoding geometry...");
		for(Vertex v: this.polyhedron.vertices) {
			float x=polyhedron.getX(v.index);
			float y=polyhedron.getY(v.index);
			float z=polyhedron.getZ(v.index);
			TC.println(x+" "+y+" "+z);
		}
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;

		System.out.println("done ("+duration+" s)");
			
		System.out.print("\tencoding connectivity...");
		startTime=System.nanoTime();
		TC.println(this.encoding.getRedWord()); // write red tree to file
		TC.println(this.encoding.getBlackWord()); // write black tree to file
		duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
		
		TC.ecritureSortieStandard();
	}

	/**
	 * Write the encoding of the triangulation to a binary file <p>
	 * <p>
	 * Warning: it works only for 3d triangle meshes of genus 0 (without boundaries). Vertices must have 3d coordinates
	 */
	public void writeToBinaryFile(String output){
		System.out.print("Writing compressed format to binary file ("+output+")");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int[] redCode=this.encoding.redWord.bits; // binary encoding of the red word
		int[] blackCode=this.encoding.blackWord.bits; // binary encoding of the black word
		int n=this.encoding.n; // number of vertices in the triangle mesh
		int totalSize=4+(3*n*4)+(redCode.length*4)+(blackCode.length*4); // total number of bytes in the binary encoding

		BinaryEncoding binary=new BinaryEncoding(totalSize);
		
		//System.out.println("\tSetting file header n="+n);
		binary.appendNextInteger(n);
		
		//System.out.print("\tEncoding vertex coordinates...");
		for(Vertex v: this.polyhedron.vertices) {
			float x=polyhedron.getX(v.index);
			float y=polyhedron.getY(v.index);
			float z=polyhedron.getZ(v.index);
			binary.appendNextFloat(x);
			binary.appendNextFloat(y);
			binary.appendNextFloat(z);
			//System.out.println(p);
		}
		//System.out.println("done");
			
		//System.out.print("\tEncoding connectivity...");
		for(int i=0;i<redCode.length;i++) { // append the red word to the binary encoding
			binary.appendNextWord(BinaryEncoding.toByteArray(redCode[i]));
		}
		for(int i=0;i<blackCode.length;i++) { // append the black word to the binary encoding
			binary.appendNextWord(BinaryEncoding.toByteArray(blackCode[i]));
		}
		//System.out.println("done");
		
		try {
			binary.writeToFile(output);
		} catch(Exception e) {}
		
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");

	}

	/**
	 * Print the orientation and coloring of all vertices
	 */
	public String printEdge(Halfedge e){
		return "("+e.getOpposite().getVertex().index+", "+e.getVertex().index+")";
	}

}
