package compactTriangulations.encoding;

import Jcg.util.BitSet32;
import arrays.ArrayBasedStack;

/**
 * @author Luca Castelli Aleardi
 * 
 * This class provides methods for dealing with the encoding of a triangulation with n vertices
 * <p>
 * The triangulation is encoded with two binary strings of length 2n. <p>
 * In this implementation the two strings are encoded separately, as two BitSets. 
 * These two strings can be written/read in a sequential manner independently.
 * 
 */
public class StringEncoderTwoWords {
	
	/** Number of vertices of the planar triangulation*/
	public final int n;

	/** Number of symbols in the red string */
	private int nRed;

	/** Number of symbols in the black string */
	private int nBlack;

	/** Bit string encoding the red word */
	public BitSet32 redWord;
	/** Bit string encoding the black word */
	public BitSet32 blackWord;
	
	// global counters for the reading and writing phases
	private int redCount=0; // global counter for red symbols
	private int blackCount=0; // global counter for black symbols
	
	/**
	 * Initialize the two bit sets
	 * 
	 * @param n  the number of vertices of the triangulation
	 */
	public StringEncoderTwoWords(int n) {
		//System.out.print("Initializing binary string decoder...");
		this.n=n;
		this.nRed=2*(n-1); // the red tree has n nodes, and 2*(n-1) edges
		this.nBlack=2*n-4; // the number of '0' bits is (n-1), since v0 is not counted; the number of black edges is (n-2)-1 
		
		this.redWord=new BitSet32(nRed);
		this.blackWord=new BitSet32(nBlack);
		
		//System.out.println(""+(this.nRed+this.nBlack)+" bits");
	}

	/**
	 * Initialize the two bit sets
	 * 
	 * @param n  the number of vertices of the triangulation
	 */
	public StringEncoderTwoWords(BitSet32 red, BitSet32 black, int n) {
		//System.out.print("Initializing binary string decoder...");
		this.n=n;
		this.nRed=2*(n-1); // the red tree has n nodes, and 2*(n-1) edges
		this.nBlack=2*n-4; // the number of '0' bits is (n-1), since v0 is not counted; the number of black edges is (n-2)-1 
		
		this.redWord=red;
		this.blackWord=black;
		//System.out.println(""+(this.nRed+this.nBlack)+" bits");
	}

	/**
	 * Re-initialize the global counters: for re-starting the reading or writing phases
	 */
	public void resetGlobalCounters() {
		this.redCount=0;
		this.blackCount=0;
	}

	/**
	 * Add a new bit to the red String
	 * 
	 * @param bit  an integer value: '0' or '1'
	 */
	public void addRedChar(int bit) {
		//System.out.println(this.redCount+" - r"+bit);
		if(bit==1) // set the bit to '1' at position 'redCount' 
			this.redWord.set(this.redCount);
		
		this.redCount++; // increment the global counter for the red word
	}

	/**
	 * Add a new bit to the black String
	 * 
	 * @param bit  an integer value: '0' or '1'
	 */
	public void addBlackChar(int bit) {
		//System.out.println("b"+bit);
		if(bit==1) // set the bit to '1' at position 'redCount' 
			this.blackWord.set(this.blackCount);
		
		this.blackCount++; // increment the global counter for the red word
	}

	/**
	 * Add a new block of (d+1) symbols to the black word: a sequence of size 'd' of one bits, followed by '0', as below:<p>
	 * '1...10'
	 * 
	 * @param degree  the number of '1' bits in the string
	 */
	public void addBlackBlock(int degree) {
		for(int i=0;i<degree;i++) { // add a sequence of '1' bits, of length 'degree'
			this.blackWord.set(this.blackCount); // set the bit to '1'
			this.blackCount++;
		}
		this.blackCount++; // increment the global counter for the last '0' in the block
	}

	/**
	 * Return the next bit in the red string <p> (for the decoding phase)
	 */
	public int readNextRedChar() {
		if(this.redCount>=this.nRed) // out of bounds (end of the bit set)
			return -1;
		
		int result=this.redWord.getBit(this.redCount);
		this.redCount++; // increment the global counter for the red word
		return result;
	}

	/**
	 * Return the black block size: in-degree of black edges. <p> (for the decoding phase) <p>
	 * <p>
	 * Remark: a block consists of a sequence of '1' bits (possibly empty) followed by a '0' bit <p>
	 * 
	 * example of a block: 0, 10, 1111110
	 * 
	 * @param return the number of '1' bits in the block (in the black word)
	 */
	public int readNextBlackBlock() {
		int degree=0; // number of '1' bits in the block
		while(this.blackCount<this.nBlack && this.blackWord.getBit(this.blackCount)==1) {
			degree++; // increment the number of '1' bits in the block
			this.blackCount++; // read next char in the black word
		}
		this.blackCount++; // scan last symbol in the block (it must be a '0' bit)
		return degree;
	}

	/**
	 * Return the two red and black strings
	 */
	public String toString() {
		String result="";
		for(int i=0;i<this.nRed;i++) {
			if(this.redWord.getBit(i)==1)
				result=result+"(";
			else
				result=result+")";
		}
		result=result+"\n";
		
		for(int i=0;i<this.nBlack;i++)
			result=result+this.blackWord.getBit(i);
		
		return result;
	}

	/**
	 * Return the two red word (encoding the red tree), as a String
	 */
	public String getRedWord() {
		char[] symbols=new char[this.nRed];
		
		for(int i=0;i<this.nRed;i++) {
			if(this.redWord.getBit(i)==1)
				symbols[i]='(';
			else
				symbols[i]=')';
		}

		return new String(symbols, 0, this.nRed);
	}

	/**
	 * Return the two red word (encoding the red tree), as a String
	 */
	public String getBlackWord() {
		char[] symbols=new char[this.nBlack];
		
		for(int i=0;i<this.nBlack;i++) {
			if(this.blackWord.getBit(i)==1)
				symbols[i]='1';
			else
				symbols[i]='0';
		}
		
		String result=new String(symbols, 0, this.nBlack);
		return result;
	}

	/**
	 * Count the max depth of the balanced red word
	 */
	public int getMaxDepthRed() {
		int max=0;
		int depth=0;
		
		for(int i=0;i<this.nRed;i++) {
			if(this.redWord.getBit(i)==1) {
				depth++;
				if(depth>max)
					max=depth;
			}
			else
				depth--;
		}
		
		return max;
	}

	/**
	 * Check the correctness of the encoding. Check a few necessary conditions:<p>
	 * <p>
	 * -) the red word must be balanced<p>
	 * -) the black word must contain exactly (n-1) '1' bits
	 */
	public boolean check() {
		System.out.print("Checking mesh encoding string: ");
		
		ArrayBasedStack stack=new ArrayBasedStack(2*(n-1)); // for checking the balanced parenthesis red word
		System.out.print("red word ");
		int countParenthesis=1;
		for(int i=0;i<this.nRed;i++) {
			if(this.redWord.getBit(i)==1) {
				stack.push(countParenthesis);
				countParenthesis++;
			}
			else
				stack.pop();
		}
		if(countParenthesis-1!=n-1) // all the vertices v1, ..., v_{n-1} must be added once to the stack
			throw new Error("Error: wrong number of vertices in the stack "+countParenthesis);
		if(stack.isEmpty()==false)
			throw new Error("Error: the word is not balanced (stack not empty)");
		System.out.print(" (ok) - ");
		
		System.out.print("black word ");
		int count1=0; // count the '1' bits in the black word
		for(int i=0;i<this.nBlack;i++)
			count1=count1+this.blackWord.getBit(i);
		if(count1!=n-3)
			throw new Error("Error: wrong number of '1' bits in the black string "+count1);
		
		System.out.println("(ok) [max depth in red word = "+this.getMaxDepthRed()+"]");
		
		return true;
	}

}
