package compactTriangulations.test;

import java.nio.file.Files;
import java.nio.file.Paths;

import Jcg.geometry.Point_3;
import Jcg.mesh.MeshBuilder;
import Jcg.mesh.MeshLoader;
import Jcg.mesh.SharedVertexRepresentation;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.schnyderwoods.PlanarTriSchnyderWood;
import compactTriangulations.CompactTriangleMesh_6nNew;
import compactTriangulations.building.CompactMeshBuilder_6nNew;
import compactTriangulations.check.CheckCorrectness;
import compactTriangulations.encoding.BinaryOFF;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2016)
 * 
 * Test the timing performances of the construction of the compact data structure CDS6n. <p>
 * Supported input formats are: <p>
 * Two types of input files are supported: <p>
 * -) standard OFF format (encoded as text file) <p>
 * -) binary OFF format: binary encoding where each integer and float value is encoded on 32 bbits (4 bytes) <p>
 * <p>
 * Use as input argument for the Java Virtual Machine: -Xms2G -Xmx2G -XX:+UseConcMarkSweepGC
 */
public class TestBuildingFromOFF {

	String filename; // name of the mesh to test
	int rootEdge;
	public CompactTriangleMesh_6nNew cdsNew6=null; // compact data structure of size 6n (new implementation)

	public TestBuildingFromOFF(String filename) {
		this.rootEdge=1; // choice of the root edge
		this.filename=filename; // input triangle mesh
	}
	
	/**
	 * Construct the CDS6n (new implementation), from a mesh stored in standard OFF format (text file)
	 */
	public void buildCDS6nFromOFF() {
		Polyhedron_3 mesh=MeshLoader.getSurfaceMesh(filename); // half-edge representation (non compact)
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_6nNew cmNew6=new CompactMeshBuilder_6nNew(cut);
		this.cdsNew6=cmNew6.buildFromPolyhedron(); // build the data structure (new implementation)
		CheckCorrectness.checkVertexDegree(this.cdsNew6);
	}

	/**
	 * Construct the CDS6n (new implementation), from a mesh stored in binary OFF format
	 */
	public void buildCDS6nFromBinaryOFF() {
		Runtime r = Runtime.getRuntime();
		r.gc();
		r.gc();
		
		PlanarTriSchnyderWood cut=this.computeSW();
		cut.performTraversal(); // computing the Schnyder orientation	

		r.gc();

		CompactMeshBuilder_6nNew cmNew6=new CompactMeshBuilder_6nNew(cut);
		this.cdsNew6=cmNew6.buildFromPolyhedron(); // build the data structure (new implementation)
		CheckCorrectness.checkVertexDegree(this.cdsNew6);
	}

	/**
	 * Construct the CDS6n (new implementation), from a mesh stored in binary OFF format and construct the Schnyder wood
	 */
	private PlanarTriSchnyderWood computeSW() {
    	Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
    	mesh.isValid(false);
    	//System.out.println(mesh.verticesToString());
    	
    	//BinaryOFF.vertexDistribution(mesh); // output the distribution of vertex degrees

		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		return cut;
	}

	/**
	 * Read an OFF file (text or binary encoding) and build the compact data structure
	 */
	public static void main(String[] args) {
		System.out.println("\n---Testing performances of compact mesh representations: construction from OFF file---\n");
		TestBuildingFromOFF test;
		
		if(args.length<1) {
			System.err.println("Error: wrong number of input parameters");
			System.err.println("One input file required: triangle mesh in standard OFF format or binary OFF format");
			System.exit(0);
		}
		else if(Files.exists(Paths.get(args[0]))==false) {
			System.err.println("Error: input file \""+args[0]+"\" not found");
			System.exit(0);
		}
		if(args[0].endsWith(".off")==true || args[0].endsWith(".OFF")==true) {
			test=new TestBuildingFromOFF(args[0]);
			test.buildCDS6nFromOFF();
		}
		else if(args[0].endsWith(".binoff")==true || args[0].endsWith(".BINOFF")==true) {
			test=new TestBuildingFromOFF(args[0]);
			
			System.out.println("\t------first run------\t");
			System.gc();
			test.buildCDS6nFromBinaryOFF();
			
			System.out.println("\n\t-------second run--------\t");
			System.gc();
			test.buildCDS6nFromBinaryOFF();
			
			System.out.println("\n\t-------third run--------\t");
			System.gc();			
			test.buildCDS6nFromBinaryOFF();
			
			System.out.println("\n\t-------fourth run--------\t");
			System.gc();
			test.buildCDS6nFromBinaryOFF();
		}
		else {
			System.out.println("Error input file (wrong format): "+args[0]);
			System.out.println("One input file required: triangle mesh in OFF format (.off) or binary OFF (.binoff)");
			System.exit(0);			
		}

	}

}
