package compactTriangulations.test;

import Jcg.polyhedron.*;
import Jcg.schnyderwoods.PlanarTriSchnyderWood;
import Jcg.mesh.MeshLoader;

import arraybasedDS.*;
import compactTriangulations.*;
import compactTriangulations.bench.EvaluateNavigationalOperations;
import compactTriangulations.building.CompactMeshBuilder_4n;
import compactTriangulations.building.CompactMeshBuilder_5n;
import compactTriangulations.building.CompactMeshBuilder_6n;
import compactTriangulations.building.CompactMeshBuilder_6nNew;
import compactTriangulations.check.*;
import compactTriangulations.encoding.BinaryOFF;

/**
 * @author Luca Castelli Aleardi
 * Comparing time performances of edge based data structures
 * Tests do not involve the use of interfaces (nor generics), in order to obtain
 * a fair comparison of data structures.
 * 
 * JVM input argument: -Xss10m
 */
public class TestVertexDegree {

	static int repeat=6; // number of times the benchmark is performed
	static double timings;
	static double average;

	String filename; // name of the mesh to test
	int rootEdge;
	
	// data structures supported for benchmarks
	public CompactTriangleMesh_6n cds6=null; // compact data structure of size 6n (old implementation: Isaac 2011 version)
	public CompactTriangleMesh_5n cds5=null; // compact data structure of size 5n (old implementation: Isaac 2011 version)
	public CompactTriangleMesh_4n cds4=null; // compact data structure of size 4n (old implementation: Isaac 2011 version)
	public CompactTriangleMesh_6nNew cdsNew6=null; // compact data structure of size 6n (new implementation)
	public ArrayBasedHalfedge hads=null; // half-edge data structure
	public ArrayBasedWingededge weds=null; // winged edge data structure
	
	public TestVertexDegree(String filename) {
		this.rootEdge=1; // choice of the root edge
		this.filename=filename; // input triangle mesh
	}

	/**
	 * Construct the CDS6n (new implementation)
	 */
	public void buildCDS6nNew() {
		//Polyhedron_3 mesh=MeshLoader.getPlanarMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_6nNew cmNew6=new CompactMeshBuilder_6nNew(cut);
		this.cdsNew6=cmNew6.buildFromPolyhedron(); // build the data structure (new implementation)
		CheckCorrectness.checkVertexDegree(this.cdsNew6);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the CDS6n (new implementation)
	 */
	public void bench6nNew() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexDegree(this.cdsNew6);
		}
		System.out.println("\t--- benchmarks done --- ("+this.cdsNew6.name()+")\n");
	}

	/**
	 * Construct the CDS6n (Isaac 2011 implementation)
	 */
	public void buildCDS6n() {
		//Polyhedron_3 mesh=MeshLoader.getPlanarMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_6n cm6=new CompactMeshBuilder_6n(cut);
		this.cds6=cm6.buildFromPolyhedron(); // build the data structure (Isaac 2011 version)
		CheckCorrectness.checkVertexDegree(this.cds6);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the CDS6n (Isaac 2011 implementation)
	 */
	public void bench6n() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexDegree(this.cds6);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.cds6.name()+")\n");
	}

	/**
	 * Construct the CDS5n (Isaac 2011 implementation)
	 */
	public void buildCDS5n() {
		//Polyhedron_3 mesh=MeshLoader.getPlanarMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_5n cm5=new CompactMeshBuilder_5n(cut);
		this.cds5=cm5.buildFromPolyhedron(); // build the data structure (Isaac 2011 version)
		CheckCorrectness.checkVertexDegree(this.cds5);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the CDS5n (Isaac 2011 implementation)
	 */
	public void bench5n() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexDegree(this.cds5);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.cds5.name()+")\n");
	}

	/**
	 * Construct the CDS4n (Isaac 2011 implementation)
	 */
	public void buildCDS4n() {
		//Polyhedron_3 mesh=MeshLoader.getPlanarMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_4n cm4=new CompactMeshBuilder_4n(cut);
		this.cds4=cm4.buildFromPolyhedron(); // build the data structure (Isaac 2011 version)
		CheckCorrectness.checkVertexDegree(this.cds4);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the CDS4n (Isaac 2011 implementation)
	 */
	public void bench4n() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexDegree(this.cds4);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.cds4.name()+")\n");
	}

	/**
	 * Construct the Half-edge data structure
	 */
	public void buildHE() {
		//Polyhedron_3 mesh=MeshLoader.getPlanarMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		
		ArrayBasedMeshBuilder mb=new ArrayBasedMeshBuilder(mesh);
		this.hads=mb.halfedgeFromPolyhedron(); // build the data structure
		CheckCorrectness.checkVertexDegree(this.hads);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the Half-edge data structure
	 */
	public void benchHE() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexDegree(this.hads);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.hads.name()+")\n");
	}

	/**
	 * Construct the Winged-edge data structure
	 */
	public void buildWE() {
		//Polyhedron_3 mesh=MeshLoader.getPlanarMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		ArrayBasedMeshBuilder mb=new ArrayBasedMeshBuilder(mesh);		
		this.weds=mb.wingededgeFromOrientedPolyhedron(cut.getEdgeOrientation()); // use the Schnyder wood orientation for orienting edges
		CheckCorrectness.checkVertexDegree(this.weds);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the Winged-edge data structure
	 */
	public void benchWE() {
		System.out.println("\t--- Starting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexDegree(this.weds);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.weds.name()+")\n");
	}

	/**
	 * Perform all tests (for each data structure)
	 */
	public static void main(String[] args) {
		System.out.println("\n---Testing performances of triangle mesh representations (vertex degree computation)---\n");
		TestVertexDegree test;
		
		if(args.length!=2) {
			System.err.println("Error: wrong number of input parameters (2 parameters required)");
			System.out.println("One input file required: triangle mesh in binary OFF format");
			System.out.println("name of the data structure to test: \n supported options are listed below");
			System.out.println("\t -HE -WE -CDS6nslow -CDS6n -CDS5n -CDS4n");
			System.exit(0);
		}
		if(args[0].endsWith(".binoff")==false) {
			System.err.println("Error input file (wrong format)");
			System.out.println("One input file required: triangle mesh in binary OFF format (.binoff)");
			System.exit(0);
		}
		
		if(args[1].equals("-HE")) {
			test=new TestVertexDegree(args[0]);
			System.gc();
			test.buildHE();
			test.benchHE();
		}

		else if(args[1].equals("-WE")) {
			test=new TestVertexDegree(args[0]);
			System.gc();
			test.buildWE();
			test.benchWE();
		}
		
		else if(args[1].equals("-CDS6nslow")) {
			test=new TestVertexDegree(args[0]);
			System.gc();
			test.buildCDS6nNew();
			test.bench6nNew();
		}
		
		else if(args[1].equals("-CDS6n")) {
			test=new TestVertexDegree(args[0]);
			System.gc();
			test.buildCDS6n();
			test.bench6n();
		}
		
		else if(args[1].equals("-CDS5n")) {
			test=new TestVertexDegree(args[0]);
			System.gc();
			test.buildCDS5n();
			test.bench5n();
		}

		else if(args[1].equals("-CDS4n")) {
			test=new TestVertexDegree(args[0]);
			System.gc();
			test.buildCDS4n();
			test.bench4n();
		}
		
		else {
			System.err.println("Error: data structure not recognized");
			System.exit(0);
		}
	}

}
