package compactTriangulations.test;

import Jcg.mesh.MeshLoader;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.schnyderwoods.PlanarTriSchnyderWood;
import arraybasedDS.ArrayBasedHalfedge;
import arraybasedDS.ArrayBasedMeshBuilder;
import arraybasedDS.ArrayBasedWingededge;
import compactTriangulations.CompactTriangleMesh_4n;
import compactTriangulations.CompactTriangleMesh_5n;
import compactTriangulations.CompactTriangleMesh_6n;
import compactTriangulations.CompactTriangleMesh_6nNew;
import compactTriangulations.bench.EvaluateNavigationalOperations;
import compactTriangulations.building.CompactMeshBuilder_4n;
import compactTriangulations.building.CompactMeshBuilder_5n;
import compactTriangulations.building.CompactMeshBuilder_6n;
import compactTriangulations.building.CompactMeshBuilder_6nNew;
import compactTriangulations.check.CheckCorrectness;
import compactTriangulations.encoding.BinaryOFF;

/**
 * @author Luca Castelli Aleardi
 * Comparing time performances of edge based data structures
 * Tests do not involve the use of interfaces (nor generics), in order to obtain
 * a fair comparison of data structures.
 * 
 * JVM input argument: -Xss10m
 */
public class TestVertexNormals {

	static int repeat=6; // number of times the benchmark is performed
	static double timings;
	static double average;

	String filename; // name of the mesh to test
	int rootEdge;
	
	// data structures supported for benchmarks
	public CompactTriangleMesh_6n cds6=null; // compact data structure of size 6n (old implementation: Isaac 2011 version)
	public CompactTriangleMesh_5n cds5=null; // compact data structure of size 5n (old implementation: Isaac 2011 version)
	public CompactTriangleMesh_4n cds4=null; // compact data structure of size 4n (old implementation: Isaac 2011 version)
	public CompactTriangleMesh_6nNew cdsNew6=null; // compact data structure of size 6n (new implementation)
	public ArrayBasedWingededge weds=null; // winged edge data structure
	
	public TestVertexNormals(String filename) {
		this.rootEdge=1; // choice of the root edge
		this.filename=filename; // input triangle mesh
	}

	/**
	 * Construct the CDS6n (new implementation)
	 */
	public void buildCDS6nNew() {
		//Polyhedron_3 mesh=MeshLoader.getSurfaceMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_6nNew cmNew6=new CompactMeshBuilder_6nNew(cut);
		this.cdsNew6=cmNew6.buildFromPolyhedron(); // build the data structure (new implementation)
		CheckCorrectness.checkVertexDegree(this.cdsNew6);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the CDS6n (new implementation)
	 */
	public void bench6nNew() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexNormal(this.cdsNew6);
		}
		System.out.println("\t--- benchmarks done --- ("+this.cdsNew6.name()+")\n");
	}

	/**
	 * Construct the CDS6n (Isaac 2011 implementation)
	 */
	public void buildCDS6n() {
		//Polyhedron_3 mesh=MeshLoader.getSurfaceMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_6n cm6=new CompactMeshBuilder_6n(cut);
		this.cds6=cm6.buildFromPolyhedron(); // build the data structure (Isaac 2011 version)
		CheckCorrectness.checkVertexDegree(this.cds6);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the CDS6n (Isaac 2011 implementation)
	 */
	public void bench6n() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexNormal(this.cds6);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.cds6.name()+")\n");
	}

	/**
	 * Construct the CDS5n (Isaac 2011 implementation)
	 */
	public void buildCDS5n() {
		//Polyhedron_3 mesh=MeshLoader.getSurfaceMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_5n cm5=new CompactMeshBuilder_5n(cut);
		this.cds5=cm5.buildFromPolyhedron(); // build the data structure (Isaac 2011 version)
		CheckCorrectness.checkVertexDegree(this.cds5);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the CDS5n (Isaac 2011 implementation)
	 */
	public void bench5n() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexNormal(this.cds5);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.cds5.name()+")\n");
	}

	/**
	 * Construct the CDS4n (Isaac 2011 implementation)
	 */
	public void buildCDS4n() {
		//Polyhedron_3 mesh=MeshLoader.getSurfaceMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		CompactMeshBuilder_4n cm4=new CompactMeshBuilder_4n(cut);
		this.cds4=cm4.buildFromPolyhedron(); // build the data structure (Isaac 2011 version)
		CheckCorrectness.checkVertexDegree(this.cds4);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the CDS4n (Isaac 2011 implementation)
	 */
	public void bench4n() {
		System.out.println("\tStarting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexNormal(this.cds4);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.cds4.name()+")\n");
	}

	/**
	 * Construct the Winged-edge data structure
	 */
	public void buildWE() {
		//Polyhedron_3 mesh=MeshLoader.getSurfaceMesh(filename); // half-edge representation (non compact)
		Polyhedron_3 mesh=BinaryOFF.fromBinaryOFFToPolyhedron(filename);
		PlanarTriSchnyderWood cut=new PlanarTriSchnyderWood(mesh, ((Halfedge) mesh.halfedges.get(this.rootEdge)).getOpposite()); // set the root face
		cut.performTraversal(); // computing the Schnyder orientation

		ArrayBasedMeshBuilder mb=new ArrayBasedMeshBuilder(mesh);		
		this.weds=mb.wingededgeFromOrientedPolyhedron(cut.getEdgeOrientation()); // use the Schnyder wood orientation for orienting edges
		CheckCorrectness.checkVertexDegree(this.weds);
	}

	/**
	 * Evaluate the performance of the vertex degree computation, for the Winged-edge data structure
	 */
	public void benchWE() {
		System.out.println("\t--- Starting benchmarks");
		for(int i=0;i<this.repeat;i++) {
			System.gc();
			EvaluateNavigationalOperations.testVertexNormal(this.weds);	
		}
		System.out.println("\t--- benchmarks done --- ("+this.weds.name()+")\n");
	}

	/**
	 * Perform all tests (for each data structure)
	 */
	public static void main(String[] args) {
		System.out.println("\n---Testing performances of triangle mesh representations (vertex normals computation)---\n");
		TestVertexNormals test;
		
		if(args.length<1 || args.length>1) {
			System.out.println("Error: wrong number of input parameters");
			System.out.println("One input file required: triangle mesh in OFF format");
			System.exit(0);
		}
		if(args[0].endsWith(".binoff")==false) {
			System.out.println("Error input file (wrong format): "+args[0]);
			System.out.println("One input file required: triangle mesh in binary OFF format (.binoff)");
			System.exit(0);
		}
		
		test=new TestVertexNormals(args[0]);
		System.gc();
		test.buildWE();
		System.gc();
		test.benchWE();
		
		test=new TestVertexNormals(args[0]);
		System.gc();
		test.buildCDS6nNew();
		System.gc();
		test.bench6nNew();
		
		test=new TestVertexNormals(args[0]);
		System.gc();
		test.buildCDS6n();
		System.gc();
		test.bench6n();
		
		test=new TestVertexNormals(args[0]);
		System.gc();
		test.buildCDS5n();
		System.gc();
		test.bench5n();
		
		test=new TestVertexNormals(args[0]);
		System.gc();
		test.buildCDS4n();
		System.gc();
		test.bench4n();
	}

}
