package compactTriangulations;

import arrays.*;
import arraybasedDS.*;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique)
 * 
 * Array based implementation of a compact data structure for triangle meshes.
 * The data structure exploits Schnyder woods (3-orientations) for triangulations
 * and requires 6n references (for triangulations having n vertices)
 * One service bit is stored in each integer (integers are encoded on 32 bits)
 */
public class CompactTriangleMesh_6n extends CompactTriangleMesh {
	
	protected TableWithTwoServiceBits table;
	
	private final int nReferences=6; // number of references per vertex
	private int nEdges;
	private int tableSize; // total number of entries in the table
	
	public CompactTriangleMesh_6n(int n) {
		this.nVertices=n;
		this.nEdges=3*n;
		this.tableSize=6*n;
		this.table=new TableWithTwoServiceBits(tableSize);

		// geometric coordinates	
		//this.coordF=new float[nVertices*3]; // for testing with simple float precision
		this.coordF=null; // the coordinates will be initialized during the construction/decompression phase
	}
	
	//--- methods for initializing the data structure --------
	public void setLeft(int e, int eLeft) { this.table.set(2*e, eLeft); }
	public void setRight(int e, int eRight) { this.table.set(2*e+1, eRight); }
	
	public void setLefServiceBit(int e) { this.table.setFirstServiceBit(2*e); }
	public void setRightServiceBit(int e) { this.table.setFirstServiceBit(2*e+1); }

	//--- methods for accessing information and navigating between mesh elements ---
	
	public int sizeOfVertices() { return this.nVertices; }
	public int sizeOfEdges() { return this.nEdges; }
	
	/** 
	 * Return the 'LeftFront' edge of the input edge 'e'.
	 * Edges are indexed with numbers between 0..3n-1 <p>
	 * <p>
	 * Every integer is encoded on 32 bits and stores two informations: <p> 
	 * -) a reference on 30 bits (the index of the edge 'e') <p>
	 * -) 2 service bits (the two rightmost bits: less significant bits)
	 * 
	 * @param e  the index of the input edge 'e'
	 **/
	public int getLeftFront(int e) {
		if(e!=e10) // check whether the edge belong to the outer face
			return this.table.elementAt(2*e); // extract the edge index from the integer on 32 bits
		
		return e20; // treat the edge (v1, v0) in a special manner (according to the Schnyder rule for the outer face)
	}

	public int getRightFront(int e) {
		if(e!=e20)
			return this.table.elementAt(2*e+1);
		return e10;
	}

	public int getLeftBack(int e) {
		if(e==e10)
			return e21;
		if(e==e21)
			return e20;
		int edgeColor=e%3;
		// the leftBack edge is oriented outgoing from v
		if(this.table.getFirstServiceBit(2*e)==0) {
			if(edgeColor==2) return e-2;
			return e+1;
		}
		// the leftBack edge is oriented toward v
		else {
			int eLeftFront=this.table.elementAt(2*e);
			if(edgeColor==0)
				return eLeftFront+2;
			return eLeftFront-1;
		}	
	}

	public int getRightBack(int e) {
		//System.out.print("rightBack "+e);
		if(e==e20)
			return e21;
		int edgeColor=e%3;

		// the rightBack edge is oriented outgoing from v
		if(this.table.getFirstServiceBit(2*e+1)==0) {
			if(edgeColor==0)
				return e+2;
			return e-1;
		}
		// the rightBack edge is oriented toward v
		else {
			int eRightFront=this.table.elementAt(2*e+1);
			int rightColor=eRightFront%3;
			if(rightColor!=edgeColor)
				return this.table.elementAt(2*eRightFront+1);
			if(rightColor==2) return eRightFront-2;
			return eRightFront+1;
		}
		
	}
	
	/** 
	 * Return the vertex source 'u' of edge 'e'=(u, v)
	 * <p>
	 * The 3 outgoing edges of vertex 'u' have numbers: 
	 * 3u, 3u+1, 3u+2
	 * 
	 * @param e  the index of an edge (a number between 0..3n-1)
	 **/
	public int getSource(int e) {
		return e/3;
	}
	
	public int getTarget(int e) {
		if(e==this.e10) return v0;
		if(e==this.e21) return v1;
		if(e==this.e20) return v0;
		int eColor=e%3;

		int eLeftFront=this.getLeftFront(e);
		int eLeftColor=eLeftFront%3;
		
		if(eLeftColor==eColor)
			return getTarget(eLeftFront);
		else
			return this.getSource(eLeftFront);
	}
	
	public int getEdge(int v) {
		if(v==v0)
			return e10;
		return v*3;
	}
	
	/** 
	 * Check whether edge e is outgoing from vertex v
	 **/
	public boolean isOutgoing(int e, int v) {
		if(e/3==v) return true;
		return false;
	}
	
	//--- navigational operations: vertex degree and vertex normal ---
	
	/** 
	 * Return the degree of a given vertex (cw direction)
	 **/
	public int vertexDegree(int v) {
		int result=1;
		int eFirst=this.getEdge(v); // edge incident to the vertex
		int pEdge; // edge turning around vertex v
		
		if(v!=v0)
			pEdge=this.getRightBack(eFirst);
		else
			pEdge=this.getLeftFront(eFirst);
		
		while(pEdge!=eFirst) {
			if(pEdge/3==v) { // the edge is outgoing from v
				pEdge=this.getRightBack(pEdge);
			}
			else
				pEdge=this.getLeftFront(pEdge);
			result++;
		}
		return result;
	}

	/** 
	 * Return the degree of a given vertex (cw direction)
	 * <p>
	 * To check
	 **/
	public int vertexDegreeFast(int v) {
		int result=1;
		int eFirst=this.getEdge(v); // edge incident to the vertex
		int pEdge; // edge turning around vertex v
		
		if(v!=v0)
			pEdge=this.getRightBack(eFirst);
		else
			pEdge=this.getLeftFront(eFirst);
		
		while(pEdge!=eFirst) {
			if(pEdge/3==v) { // the edge is outgoing from v (TO CHECK)
				pEdge=this.getRightBack(pEdge);
			}
			else
				pEdge=this.getLeftFront(pEdge);
			result++;
		}
		return result;
	}

	/**
	 * Return the normal to a vertex in the mesh (simple float precision)
	 * Fast version with simple Float precision, with no use of Point_ interface
	 */
	public float[] getNormalFloat(int v) {
		float[] result=new float[3]; // average of cross product vectors
		float[] faceNormal; // cross product vector
		int degree=1; // vertex degree
		int u, w; // indices of vertices u and v
		float[] pV, pU, pW; // coordinates of vertices v, u and w

		int edge=this.getEdge(v); 	// edge incident to vertex v
		int pEdge;					// edge turning around vertex v
		if(this.getSource(edge)==v) { // the edge is outgoing from v
			pEdge=this.getRightBack(edge);
			u=this.getTarget(edge);
			v=this.getSource(edge);
		}
		else { // the edge is ingoing (having v as target vertex)
			pEdge=this.getLeftFront(edge);
			v=this.getTarget(edge);
			u=this.getSource(edge);
		}
		if(this.getSource(pEdge)==v) // pEdge is outgoing from v
			w=this.getTarget(pEdge);
		else // pEdge is ingoing (having v as target vertex)
			w=this.getSource(pEdge);

		pV=this.getPoint(v); // coordinates of vertex v
		pU=this.getPoint(u); // coordinates of vertex u
		pW=this.getPoint(w); // coordinates of vertex w
		
		result=NavigationInterface.computeFaceNormal(pU, pV, pW);
		pU=pW;
		
		while(pEdge!=edge) {
			if(this.getSource(pEdge)==v) {
				pEdge=this.getRightBack(pEdge);
			}
			else {
				pEdge=this.getLeftFront(pEdge);
			}
			if(this.getSource(pEdge)==v) // pEdge is outgoing from v
				w=this.getTarget(pEdge);
			else // pEdge is ingoing (having v as target vertex)
				w=this.getSource(pEdge);
			
			pW=this.getPoint(w); // updating coordinates of vertex w	
			
			faceNormal=NavigationInterface.computeFaceNormal(pU, pV, pW);

			result[0]=result[0]+faceNormal[0];
			result[1]=result[1]+faceNormal[1];
			result[2]=result[2]+faceNormal[2];

			pU=pW;
			degree++;
		}	
		result[0]=result[0]/degree;
		result[1]=result[1]/degree;
		result[2]=result[2]/degree;
		
		return result;			
	}
		
	//--- methods for visualizing the data structure ---
	
	/**
	 * Return a String which represents the reference table encoding of the data structure
	 */
	public String toString() {
		String result="Compact Triangle Mesh DS 6n (Isaac 2011 version)\n";
		
		for(int i=0;i<this.nVertices;i++) {
			result=result+"v "+i+":";
			for(int j=0;j<this.nReferences;j++) {
				result=result+"\t"+this.table.elementToString(i*this.nReferences+j);
			}
			result=result+"\n";
		}
		
		return result;
	}
	
	/**
	 * Return the name of the data structure
	 */
	public String name() {
		return "CDS6n";
	}

}
