package compactTriangulations;

import arrays.TableWithTwoServiceBits;
import compactTriangulations.test.MemoryBenchmark;
import arraybasedDS.NavigationInterface;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2016)
 * 
 * Array based implementation of a compact data structure for triangle meshes.
 * The data structure exploits Schnyder woods (3-orientations) for triangulations
 * and requires 6n references (for triangulations having n vertices) <p>
 * <p>
 * In this implementation each edge 'e' is represented as a pair e=(u, color) <p>
 * -) the index 'u' of the source vertex, which is stored on 30 bits (the most significant bits) <p>
 * -) the 'color' of the edge (0, 1, or 2) which is stored on 2 bits (the less significant bits, on the right)
 */
public class CompactTriangleMesh_6nNew extends CompactTriangleMesh {
	
	/** The table storing the whole data structure. <p> All the arrays U (integer references on 30 bits), T (booleans) and S (booleans) are stored 
	 * in an unique integer array of size 6*n (having n rows, 6 columns). First two columns
	 * store 'red' edges; columns 2 and 3 store 'blue' edges; last two columns store 'black' edges **/
	public TableWithTwoServiceBits table;
	
	private final int nReferences=6; // number of references per vertex
	private int nEdges;
	private int tableSize; // total number of entries in the table
	
	public CompactTriangleMesh_6nNew(int n) {
		this.nVertices=n;
		this.nEdges=3*n;
		this.tableSize=6*n;
		
		this.table=new TableWithTwoServiceBits(tableSize);
		
		// geometric coordinates		
		//this.coordF=new float[nVertices*3]; // for testing with simple float precision
		this.coordF=null; // the coordinates will be initialized during the construction/decompression phase
	}
	
	//--- methods for initializing the data structure (called only during the building phase) --------
	
	/** 
	 * Store the reference in U_{color}^{left or right}[v], for an edge outgoing from v having a given color
	 * <p>
	 * The six arrays U are stored in an unique integer array of size 6*n (having n rows, 6 columns). First two columns
	 * store 'red' edges; columns 2 and 3 store 'blue' edges; last two column store 'black' edges
	 * 
	 * @param v  the index of the vertex (gives the position in the table: references are to be stored at row number 'v')
	 * @param vertexRef  the reference to be stored (an integer in {0..n-1}
	 * @param color the color of the neighboring edge (incident to v): an integer in {0, 1, 2}
	 * @param leftRight  an integer {0, 1} specifying the left or right column (of a given color)
	 **/
	public void setU(int v, int vertexRef, int color, int leftRight) {
		if(v<0 || v>=this.nVertices)
			throw new Error("Error: wrong index v"+v);
		if(vertexRef<0 || vertexRef>=this.nVertices)
			throw new Error("Error: wrong reference "+vertexRef);
		if(color<0 || color>2)
			throw new Error("Error: wrong edge color "+color);
		if(leftRight<0 || leftRight>1)
			throw new Error("Error: wrong column");
		
		int index=6*v+2*color+leftRight;
		//this.table.set(index, vertexRef); // with this instruction the two service bits are deleted (set to '0')
		//this.table.elements[index]=this.table.elements[index]+(vertexRef<<2); // service bits are preserved, but it does not work is the element is already set
		
		// the instructions below allow to set the element correctly: service bits are preserved, the element can be updated
		int firstBit=this.table.getFirstServiceBit(index);
		int secondBit=this.table.getSecondServiceBit(index);
		this.table.elements[index]=1*firstBit+2*secondBit+(vertexRef<<2); // this instruction allow to preserve the two service bits
	}

	/** Set the bit in T_{color}^{left or right}[v] equal to '1', for an edge outgoing from v having a given color
	 * <p>
	 * The six ('booleans') arrays T are stored as 'service bits' in an unique integer array of size 6*n (having n rows, 6 columns). 
	 * First two columns deal with 'red' edges; columns 2 and 3 store service bits of 'blue' edges; last two column deal with 'black' edges
	 * 
	 * @param v  the index of the vertex (gives the position in the table: references are to be stored at row number 'v')
	 * @param color the color of the neighboring edge (incident to v): an integer in {0, 1, 2}
	 * @param leftRight  an integer {0, 1} specifying the left or right column (of a given color)
	 **/
	public void setT(int v, int color, int leftRight) { 
		if(v>=this.nVertices)
			throw new Error("Error: wrong index number "+v);
		if(color<0 || color>2)
			throw new Error("Error: wrong edge color "+color);
		if(leftRight<0 || leftRight>1)
			throw new Error("Error: wrong column");
		
		if(this.getT(v, color, leftRight)==1) // nothing to do the bit is already set
			return;
		
		this.table.setFirstServiceBit(6*v+2*color+leftRight);
	}

	/** Set the bit in S_{color}[v] equal to '1', for an edge outgoing from v having a given color
	 * <p>
	 * The three ('booleans') arrays S are stored as 'service bits' in an unique integer array of size 6*n (having n rows, 6 columns).
	 * Service bits are stored in column 0, 2, 4, respectively for red, blue and black edges.
	 * <p>
	 * Remark: we have to use the 'second service bit', since the first service bit is already used for storing the booleans of tables T
	 * 
	 * @param v  the index of the vertex (gives the position in the table: references are to be stored at row number 'v')
	 * @param color the color of the neighboring edge (incident to v): an integer in {0, 1, 2}
	 **/
	public void setS(int v, int color) { 
		if(v>=this.nVertices)
			throw new Error("Error: wrong index number "+v);
		if(color<0 || color>2)
			throw new Error("Error: wrong edge color "+color);
		
		if(this.getS(v, color)==1) // nothing to do: the bit is already set
			return;
		
		this.table.setSecondServiceBit(6*v+2*color); // the bit in S[v] is stored as second service bits (since the first service bit is already used)
	}

	/** Set the bit in S_{color}[v] equal to '1', for an edge outgoing from v having a given color
	 * <p>
	 * The three ('booleans') arrays S are stored as 'service bits' in an unique integer array of size 6*n (having n rows, 6 columns).
	 * Service bits are stored in column 0, 2, 4, respectively for red, blue and black edges.
	 * <p>
	 * Remark: we have to use the 'second service bit', since the first service bit is already used for storing the booleans of tables T
	 * 
	 * @param v  the index of the vertex (gives the position in the table: references are to be stored at row number 'v')
	 * @param color the color of the neighboring edge (incident to v): an integer in {0, 1, 2}
	 **/
	public void setSFalse(int v, int color) { 
		if(v>=this.nVertices)
			throw new Error("Error: wrong index number "+v);
		if(color<0 || color>2)
			throw new Error("Error: wrong edge color "+color);
		
		if(this.getS(v, color)==0) // alreay set to '0': nothing to do
			return;
		
		//this.table.setSecondServiceBit(6*v+2*color); // the bit in S[v] is stored as second service bits (since the first service bit is already used)
		int index=6*v+2*color;
		this.table.elements[index]=this.table.elements[index]-2;
	}

	/** Set the geometric coordinates of vertex 'v'
	 **/
	public void setPoint(int v, float x, float y, float z) {
		int index=3*v;
		this.coordF[index]=x;
		this.coordF[index+1]=y;
		this.coordF[index+2]=z;			
	}
	
	//--- low-level methods for accessing information in tables S, T and U ---
	/** 
	 * Retrieve the reference in U_{color}^{left or right}[v], for an edge outgoing from v having a given color
	 * <p>
	 * The six arrays U are stored in an unique integer array of size 6*n (having n rows, 6 columns). First two columns
	 * store 'red' edges; columns 2 and 3 store 'blue' edges; last two column store 'black' edges
	 * 
	 * @param v  the index of the vertex (gives the position in the table: references are to be stored at row number 'v')
	 * @param color the color of the neighboring edge (incident to v): an integer in {0, 1, 2}
	 * @param leftRight  an integer {0, 1} specifying the left or right column (of a given color)
	 **/
	public int getU(int v, int color, int leftRight) {
		return this.table.elementAt(6*v+2*color+leftRight);
	}

	/** Set the bit in T_{color}^{left or right}[v] equal to '1', for an edge outgoing from v having a given color
	 * <p>
	 * The six ('booleans') arrays T are stored as 'service bits' in an unique integer array of size 6*n (having n rows, 6 columns). 
	 * First two columns deal with 'red' edges; columns 2 and 3 store service bits of 'blue' edges; last two column deal with 'black' edges
	 * 
	 * @param v  the index of the vertex (gives the position in the table: references are to be stored at row number 'v')
	 * @param color the color of the neighboring edge (incident to v): an integer in {0, 1, 2}
	 * @param leftRight  an integer {0, 1} specifying the left or right column (of a given color)
	 **/
	public int getT(int v, int color, int leftRight) { 
		return this.table.getFirstServiceBit(6*v+2*color+leftRight);
	}

	/** Set the bit in S_{color}[v] equal to '1', for an edge outgoing from v having a given color
	 * <p>
	 * The three ('booleans') arrays S are stored as 'service bits' in an unique integer array of size 6*n (having n rows, 6 columns).
	 * Service bits are stored in column 0, 2, 4, respectively for red, blue and black edges.
	 * <p>
	 * Remark: we have to use the 'second service bit', since the first service bit is already used for storing the booleans of tables T
	 * 
	 * @param v  the index of the vertex (gives the position in the table: references are to be stored at row number 'v')
	 * @param color the color of the neighboring edge (incident to v): an integer in {0, 1, 2}
	 **/
	public int getS(int v, int color) {
		return this.table.getSecondServiceBit(6*v+2*color); // the bit in S[v] is stored as second service bit
	}

	//--- methods for navigating between mesh elements ---
	
	public int sizeOfVertices() { return this.nVertices; }
	public int sizeOfEdges() { return this.nEdges; }
	
	/** 
	 * Return the 'LeftFront' edge of the input edge 'e=(u, v)' having color 'c'.
	 * The source vertex 'u' is an index between 0..n-1.
	 * The edge color 'c' is encoded on 2 service bits.
	 * 
	 * @param e  an integer (on 32 bits) encoding the pair (u, c)
	 **/
	public int getLeftFront(int e) {
		if(e!=e21) { // check whether the edge (u,v) must be treated in a special manner
			int eSource=this.getSource(e); // retrieve the source of edge 'e'
			int eColor=this.getColor(e); // retrieve the color of edge 'e'
			int eLeftFront=this.getU(eSource, eColor, 0); // access the array 'U' storing the reference of the leftfront edge
			
			int colorLeftFront=eColor;
			
			// read the boolean stored in array 'T' 
			if(this.getT(eSource, eColor, 0)==0) // edges 'e' and 'LeftFront' must have different colors
				colorLeftFront=nextColor(eColor); // compute the color of the leftfront edge (according to Schnyder local rule)
			
			// convert the pair (e, c) into an integer value (on 32 bits)
			return TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(eLeftFront, colorLeftFront);
		}
		return e10;
	}

	/** 
	 * Return the 'RightFront' edge, neighbor of e=(u, color)
	 * 
	 * @param e  an integer encoding (u, color)
	 **/
	public int getRightFront(int e) {
		if(e!=e20) {
			int eSource=this.getSource(e);
			int eColor=this.getColor(e);
			int eRightFront=this.getU(eSource, eColor, 1);
			
			int colorRightFront=eColor;
			if(this.getT(eSource, eColor, 1)==0) // edges 'e' and 'RightFront' have different colors
				colorRightFront=this.prevColor(eColor);
			
			return TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(eRightFront, colorRightFront);
		}
		return e10;
	}

	/** 
	 * Return the 'LeftBack' edge, neighbor of e=(u, color)
	 * 
	 * @param e  an integer encoding (u, color)
	 * @return  an integer encoding the 'LeftBack' edge: (source, color)
	 **/
	public int getLeftBack(int e) {
		if(e==e10)
			return e21;
		if(e==e21)
			return e20;
		int edgeColor=e%3;
		// the leftBack edge is oriented outgoing from v
		if(this.table.getFirstServiceBit(2*e)==0) {
			if(edgeColor==2) return e-2;
			return e+1;
		}
		// the leftBack edge is oriented toward v
		else {
			int eLeftFront=this.table.elementAt(2*e);
			if(edgeColor==0)
				return eLeftFront+2;
			return eLeftFront-1;
		}	
	}

	/** 
	 * Return the 'RightBack' edge, neighbor of e=(u, color)
	 * 
	 * @param e  an integer encoding (u, color)
	 * @return  an integer encoding the 'RightBack' edge: (source, color)
	 **/
	public int getRightBack(int e) {
		//System.out.print("rightBack "+e);
		if(e==e20)
			return e21;
		int eColor=this.getColor(e);
		int eSource=this.getSource(e);

		// Case 1: the rightBack edge (u, z) is oriented outgoing from 'u'
		if(this.getS(eSource, this.nextColor(eColor))==0) {
			//System.out.print(" [Case 1: color"+this.nextColor(eColor)+"] ");
			return TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(eSource, this.prevColor(eColor));
		}
			
		// Cases 2 and 3: the rightBack edge (u, z) is oriented toward u
		else {
			int eRightFront=this.getRightFront(e); // right front edge (v, z)
			int isSameColor=this.getT(eSource, eColor, 1); // check whether edges (u, v) and (v, z) have the same color
			if(isSameColor==1) { // Case 2: edge (z, v) is oriented toward v
				//System.out.print(" [Case 2] ");
				return TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(this.getSource(eRightFront), this.nextColor(eColor));
			}
			
			else { // Case 3: edge (v, z) is oriented toward z
				//System.out.print(" [Case 3] ");
				return this.getRightFront(eRightFront);
			}
		}
		
	}

	/** 
	 * Return the vertex source 'u' of edge 'e'=(u, color)
	 * <p>
	 * By definition, just perform a bit shift (by 2)
	 * 
	 * @param e  an integer encoding (u, color)
	 **/
	public int getSource(int e) {
		return (e>>2);
	}

	/** 
	 * Retrieve the color of edge 'e'=(u, color)
	 * <p>
	 * By definition, the color is stored in the last 2 (right-most) bits of the integer
	 * 
	 * @param e  an integer encoding (u, color)
	 **/
	public int getColor(int e) {
		return e & (0x01+0x02); // the mask '(0x01+0x02)' means 00000011
	}

	/** 
	 * Retrieve the target vertex of edge 'e'=(u, color)
	 * <p>
	 * By definition, the color is stored in the last 2 (right-most) bits of the integer
	 * 
	 * @param e  an integer encoding (u, color)
	 **/
	public int getTarget(int e) {
		if(e==this.e10) return v0;
		if(e==this.e21) return v1;
		if(e==this.e20) return v0;
		int eColor=this.getColor(e);

		int eLeftFront=this.getLeftFront(e);
		int eLeftColor=this.getColor(eLeftFront);
		
		if(eLeftColor==eColor)
			return getTarget(eLeftFront);
		else
			return this.getSource(eLeftFront);
	}

	/** Return an edge incident to vertex v
	 * <p>
	 * By default, return the edge of color red: (v, 0) for all vertices except v0
	 * 
	 * @param v  the index of the vertex
	 **/
	public int getEdge(int v) {
		if(v==this.v0)
			return TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(this.v1, 0); // for vertex v0, return (v1, 0)
		return TableWithTwoServiceBits.storeReferenceAndServiceBitsIntoInteger(v, 0); // return the edge (v, 0)
	}
		
	/** 
	 * Check whether the edge 'e' is outgoing from vertex 'v'
	 * 
	 * @param e  an integer encoding an edge (u, color)
	 * @param v  the index of a vertex
	 * @return return true whether 'u'=='v'
	 **/
	public boolean isOutgoing(int e, int v) {
		if(v==this.v0)
			return false; // as there are not edges outgoing from v0
		
		int eSource=TableWithTwoServiceBits.extractReferenceFromInteger(e); // get the source of edge e
		return eSource==v; // test whether v coincides with the source of e
	}
	
	//--- navigational methods: vertex degree and vertex normals ---

	/** 
	 * Return the degree of a given vertex
	 **/
	public int vertexDegree(int v) {
		int result=1;
		int eFirst=this.getEdge(v); // edge incident to the vertex
		int pEdge; // edge turning around vertex v
		
		if(this.isOutgoing(eFirst, v)==true)
			pEdge=this.getRightBack(eFirst);
		else
			pEdge=this.getLeftFront(eFirst);
		
		while(pEdge!=eFirst) {
			if(this.isOutgoing(pEdge, v)==true) {
				pEdge=this.getRightBack(pEdge);
			}
			else
				pEdge=this.getLeftFront(pEdge);
			result++;
		}
		return result;
	}
	
	/**
	 * Compute the normal to a vertex in the mesh, with simple float precision<p>
	 * (neighbors are traversed turning in cw direction)
	 * 
	 * @param v  the index of vertex
	 * @return the vertex normal (in 3D)
	 */
	public float[] getNormalFloat(int v) {
		float[] result=new float[3]; // average of cross product vectors
		float[] faceNormal; // cross product vector
		int degree=1; // vertex degree
		int u, w; // indices of vertices u and v
		float[] pV, pU, pW; // coordinates of vertices v, u and w

		int edge=this.getEdge(v); 	// edge incident to vertex v
		int pEdge;					// edge turning around vertex v
		if(this.getSource(edge)==v) { // the edge is outgoing from v
			pEdge=this.getRightBack(edge);
			u=this.getTarget(edge);
			v=this.getSource(edge);
		}
		else { // the edge is ingoing (having v as target vertex)
			pEdge=this.getLeftFront(edge);
			v=this.getTarget(edge);
			u=this.getSource(edge);
		}
		if(this.getSource(pEdge)==v) // pEdge is outgoing from v
			w=this.getTarget(pEdge);
		else // pEdge is ingoing (having v as target vertex)
			w=this.getSource(pEdge);

		pV=this.getPoint(v); // coordinates of vertex v
		pU=this.getPoint(u); // coordinates of vertex u
		pW=this.getPoint(w); // coordinates of vertex w
		
		result=NavigationInterface.computeFaceNormal(pU, pV, pW);
		pU=pW;
		
		while(pEdge!=edge) {
			if(this.getSource(pEdge)==v) {
				pEdge=this.getRightBack(pEdge);
			}
			else {
				pEdge=this.getLeftFront(pEdge);
			}
			if(this.getSource(pEdge)==v) // pEdge is outgoing from v
				w=this.getTarget(pEdge);
			else // pEdge is ingoing (having v as target vertex)
				w=this.getSource(pEdge);
			
			pW=this.getPoint(w); // updating coordinates of vertex w	
			
			faceNormal=NavigationInterface.computeFaceNormal(pU, pV, pW);

			result[0]=result[0]+faceNormal[0];
			result[1]=result[1]+faceNormal[1];
			result[2]=result[2]+faceNormal[2];

			pU=pW;
			degree++;
		}	
		result[0]=result[0]/degree;
		result[1]=result[1]/degree;
		result[2]=result[2]/degree;
		
		return result;			
	}

	//--- methods for visualizing the data structure ---
	
	/**
	 * Return a String which represents the reference table encoding the data structure
	 */
	public String toString() {
		String result="Compact Triangle Mesh DS (6n references)\n";
		
		for(int i=0;i<this.nVertices;i++) {
			result=result+"v "+i+":";
			for(int j=0;j<this.nReferences;j++) {
				result=result+"\t"+this.table.elementToString(i*this.nReferences+j);
			}
			result=result+"\n";
		}
		
		return result;
	}
	
	/**
	 * Return a String representing an edge e=(u, color)
	 */
	public String printEdge(int e) {
		int u=TableWithTwoServiceBits.extractReferenceFromInteger(e);
		int color=TableWithTwoServiceBits.extractTwoServiceBitsFromInteger(e);	
		
		return "(v"+u+", "+color+")";
	}

	/**
	 * Check whether an edge e=(u, color) is valid (useful for checking correctness of the data structure)
	 * <p>
	 * Remark: <p>
	 * -) all inner edges are valid <p>
	 * -) on the outer face: only the edges e10, e20 and e21 are valid
	 */
	public boolean isEdge(int e) {
		if(e==e10 || e==e20 || e==e21) // edge 'e' belongs to the outer face
			return true;
		
		int u=this.getSource(e);
		if(u!=v0 && u!=v1 && u!=v2) // edge 'e' is incident to an inner vertex
			return true;
		
		return false; // the pair (u, color) does not define a valid edge
	}

	/**
	 * Return the name of the data structure
	 */
	public String name() {
		return "CDS6n slow";
	}

}
